 /**----------------------------------------------------
 * File:      SVGImporter.java
 * Package:   at.hoertler.svgeditor.common
 * Project:   SVGEditor
 * ----------------------------------------------------
 * @author Philipp Hörtler BSc
 * @version 1.0.0
 * Date: 2013-03-24
 * ----------------------------------------------------
 * Description:
 *
 * The SVGImporter is used to parse SVG-Code.
 * It can handle SVG-Files and PNG/PDF if they contain the 
 * neccessary SVG-Code as Metadata.
 */
package at.hoertler.svgeditor.common;

import at.hoertler.svgeditor.options.OptionManager;
import at.hoertler.svgeditor.svg.elements.SVGArrow;
import at.hoertler.svgeditor.svg.elements.SVGComposite;
import at.hoertler.svgeditor.svg.elements.SVGElement;
import at.hoertler.svgeditor.svg.elements.SVGEllipse;
import at.hoertler.svgeditor.svg.elements.SVGLine;
import at.hoertler.svgeditor.svg.elements.SVGPolygone;
import at.hoertler.svgeditor.svg.elements.SVGRectangle;
import at.hoertler.svgeditor.svg.elements.SVGText;
import at.hoertler.svgeditor.svg.elements.SVGTriangle;
import com.itextpdf.text.pdf.PdfReader;
import java.awt.Color;
import java.awt.Font;
import java.awt.Point;
import java.awt.Rectangle;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.Vector;
import javax.swing.JFileChooser;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;


public class SVGImporter {


  public SVGImporter() {
  }

  private static JFileChooser createFileChooser(boolean openDialog) {
    JFileChooser fileChooser = new JFileChooser();
    fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
    fileChooser.addChoosableFileFilter(new FileNameExtensionFilter("Portable Network Graphics", "png"));
    fileChooser.setFileFilter(new FileNameExtensionFilter("Portable Document Format", "pdf"));
    fileChooser.addChoosableFileFilter(new FileNameExtensionFilter("Scaleable Vector Graphics", "svg"));

    int ret = -500;
    if (openDialog) {
      fileChooser.setDialogType(JFileChooser.OPEN_DIALOG);
      ret = fileChooser.showOpenDialog(null);
    } else {
      fileChooser.setDialogType(JFileChooser.SAVE_DIALOG);
      ret = fileChooser.showSaveDialog(null);
    }
    //fileChooser.setVisible(true);
    if(ret == JFileChooser.APPROVE_OPTION) {
      File f = fileChooser.getSelectedFile();
      if(!fileChooser.getFileFilter().accept(f)) {
        if(fileChooser.getFileFilter().getClass().isAssignableFrom(FileNameExtensionFilter.class)) {
          fileChooser.setSelectedFile(new File(fileChooser.getSelectedFile() + "." + ((FileNameExtensionFilter)fileChooser.getFileFilter()).getExtensions()[0]));
        }
      }
      return fileChooser;
    }
    
    return null;
  }

  public static File LoadOpenFile() {
    JFileChooser fc = SVGImporter.createFileChooser(true);
    if(fc != null)
      return fc.getSelectedFile();
    return null;
  }

  public static File LoadSaveFile() {
    JFileChooser fc = SVGImporter.createFileChooser(false);
    
    if(fc != null) {
      
      return fc.getSelectedFile();
    }
    return null;
  }

  public static File LoadFile(String filepath) {
    return new File(filepath);
  }

  public static Vector<SVGElement> Import() {
    return SVGImporter.Import(SVGImporter.createFileChooser(true).
            getSelectedFile());
  }

  public static Vector<SVGElement> Import(String filepath) {
    return SVGImporter.Import(new File(filepath));
  }

  public static Vector<SVGElement> Import(File file) {
    Vector<SVGElement> elements = new Vector<SVGElement>();
    if (!file.exists()
            || !file.isFile()
            || !file.canRead()) {
      return elements;
    }
    if(file.getName().substring(file.getName().lastIndexOf('.') + 1).equalsIgnoreCase("pdf")) {
      try {
        PdfReader reader = new PdfReader(file.getPath());
        Map<String, String> info = reader.getInfo();
        String svgInfo = info.get("unknown");
        if(svgInfo == null || svgInfo.isEmpty() || !svgInfo.startsWith("<svg "))
          return elements;
        file = File.createTempFile("tmpImportFile", "svg");
        
        BufferedWriter outWr = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file), "UTF8"));
        //FileWriter outFile = new FileWriter(file);
        //PrintWriter out = new PrintWriter(outFile);
        PrintWriter out = new PrintWriter(outWr);
        out.print(svgInfo);
        out.close();
        reader.close();
      } catch(IOException ex) {

      }
    } else if(file.getName().substring(file.getName().lastIndexOf('.') + 1).equalsIgnoreCase("png")) {
      try {
        String svgInfo = HelperFunction.getPNGMetaData(file, "SVGCode");
        if(svgInfo == null || svgInfo.isEmpty() || !svgInfo.startsWith("<svg "))
          return elements;
        file = File.createTempFile("tmpImportFile", "svg");

        BufferedWriter outWr = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file), "UTF8"));
        //FileWriter outFile = new FileWriter(file);
        //PrintWriter out = new PrintWriter(outFile);
        PrintWriter out = new PrintWriter(outWr);
        out.print(svgInfo);
        out.close();
      } catch(IOException ex) {

      }
    } else if(!file.getName().substring(file.getName().lastIndexOf('.') + 1).equalsIgnoreCase("svg"))
      return elements;
      
    

    try {
      DocumentBuilder db = DocumentBuilderFactory.newInstance().newDocumentBuilder();
      Document doc = db.parse(file);
      doc.getDocumentElement().normalize();
      NodeList nList = doc.getChildNodes();
      if (nList.getLength() > 0 && nList.item(0).getNodeType() == Node.ELEMENT_NODE)
        elements = importElements(((Element) nList.item(0)).getChildNodes());
    } catch (Exception ex) {
      System.err.println("Import Error: import aborted due to exception: " + ex.toString());
    }
    

    return elements;
  }

  private static Vector<SVGElement> importElements(NodeList nodes)
          throws XPathExpressionException {
    Vector<SVGElement> elements = new Vector<SVGElement>();
    SVGElement validElem = null;
    for (int i = 0; i < nodes.getLength(); i++) {
      Node n = nodes.item(i);
      if (!n.getNodeName().equals("#text") && !n.getNodeName().equals("#comment")) {
        if (n.getChildNodes() != null && n.getChildNodes().getLength() > 0) 
          validElem = importSVGElement((Element)n);

        if(validElem!=null && validElem.validate())
          elements.add(validElem);
        else
          System.err.println("Import Error: Validation failed on " + validElem + ".");
      }
      validElem = null;
    }
    return elements;
  }
  
  private static SVGElement importSVGElement(Element elem) throws XPathExpressionException {
    SVGElement impElem = null;
    String g_id = elem.getAttribute("id");
    
    if (g_id.equalsIgnoreCase("SVGComposite")) {
      SVGComposite comp = new SVGComposite();
      comp.setElements(importElements(elem.getChildNodes()));
      if(!comp.getElements().isEmpty())
        impElem = comp;
    } else if(g_id.equalsIgnoreCase("SVGLine")) {
      impElem = importSVGLine(elem);
    } else if(g_id.equalsIgnoreCase("SVGRoundSideRectangle")) {
      impElem = importSVGRoundSidedRectangle(elem);
    } else if(g_id.equalsIgnoreCase("SVGPolygone")) {
      impElem = importSVGPolygone(elem);
    } else if(g_id.equalsIgnoreCase("SVGText")) {
      impElem = importSVGText(elem);
    } else if(g_id.equalsIgnoreCase("SVGRectangle")) {
      impElem = importSVGRectangle(elem);
    } else if(g_id.equalsIgnoreCase("SVGEllipse")) {
      impElem = importSVGEllipse(elem);
    } else if(g_id.equalsIgnoreCase("SVGTriangle")) {
      impElem = importSVGTriangle(elem);
    }
    return impElem;
  }

  private static SVGElement importElement(Element eElement)
          throws XPathExpressionException {
    switch (eElement.getNodeName()) {
      case "line":
        return importSVGLine(eElement);
      case "rect":
        return importSVGRectangle(eElement);
      case "text":
        return importSVGText(eElement);
      case "ellipse":
        return importSVGEllipse(eElement);
      case "SVGComposite":
        SVGComposite comp = new SVGComposite();

        return comp;
      default:
        return null;
    }
  }

  private static SVGElement importSVGLine(Element element)
          throws XPathExpressionException {
    //this.printSVGElement(eElement);
    SVGLine elem = new SVGLine();

    Node n = element;
    Node textNode = null;
    Node lineNode = null;
    Node pathNode1 = null;
    Node pathNode2 = null;
    
    for(int i = 0; i<element.getChildNodes().getLength(); i++) {
      n = element.getChildNodes().item(i);
      if(n.getNodeName().equalsIgnoreCase("text"))
        textNode = n;
      else if(n.getNodeName().equalsIgnoreCase("line"))
        lineNode = n;
      else if(n.getNodeName().equalsIgnoreCase("path")) {
        if(pathNode1==null)
          pathNode1 = n;
        else
          pathNode2 = n;
      }
    }
    
    if(lineNode!=null) {
      elem = SVGImporter.importLine(lineNode);
    }
    
    if (pathNode1!=null) {
      SVGImporter.setArrow(elem, elem.getHandles().get(0), elem.getHandles().get(1), pathNode1);
    }
    
    if (pathNode2!=null) {
      SVGImporter.setArrow(elem, elem.getHandles().get(0), elem.getHandles().get(1), pathNode2);
    }
    
    if (textNode!=null) {
      if(textNode!=null) {
        elem.setMode(SVGElement.Mode.Edit);
        elem.setMode(SVGElement.Mode.TextEdit);
        SVGText text = SVGImporter.importText(textNode, elem.getBoundingBox(), false);
        elem.setText(text.getText());
        elem.setTextColor(text.getTextColor());
        elem.setTextAlignment(text.getTextAlignment());
        elem.setTextSize(text.getTextSize());
        elem.setTextOption(text.getTextOption());
        elem.setFont(text.getFont());
      }
      elem.setMode(SVGElement.Mode.Edit);
      elem.setMode(SVGElement.Mode.Normal);
    } 
     
    return elem;
  }
  
  private static SVGElement importSVGTriangle(Element elem) {
    SVGTriangle tri = new SVGTriangle();
    Node n = null;
    for(int i = 0; i<elem.getChildNodes().getLength(); i++) {
      n = elem.getChildNodes().item(i);
      if(n.getNodeName().equalsIgnoreCase("path")) {
          i++;
          break;
        }
    }
    if(n==null)
      return tri;

    Vector<Point> pts = new Vector<Point>();
    for(PathPoint pp : SVGImporter.getPathPointsForAttributeName("d", n)) {
      switch(pp.type) {
        case MoveToAbsolute:
          pts.add(pp.points[0]);
           break;
        case LineToAbsolute:
          pts.add(pp.points[0]);
          break;
      }
    }
    tri.setHandles(pts);
    
    tri.setLineStrength(SVGImporter.getLineStrength(n));
    tri.setLineColor(SVGImporter.getRGBColorForAttributeName("stroke", n));
    tri.setFillColor(SVGImporter.getRGBColorForAttributeName("fill", n));

    tri.setMode(SVGElement.Mode.Normal);
    tri.deSelect();
    return tri;
  }
  
  
  private static SVGElement importSVGPolygone(Element element) {
    SVGPolygone poly = new SVGPolygone();
    Node n = null;
    int i = 0;
    for(; i<element.getChildNodes().getLength(); i++) {
      n = element.getChildNodes().item(i);
      if(n.getNodeName().equalsIgnoreCase("path")) {
          i++;
          break;
        }
    }
    if(n==null)
      return poly;
    poly.getHandles().clear();
    Point start = new Point(0,0);
    Point end = new Point(0,0);
    Vector<PathPoint> pathPoints = SVGImporter.getPathPointsForAttributeName("d", n);
    for(PathPoint pp : pathPoints) {
      switch(pp.type) {
        case MoveToAbsolute:
          poly.addHandle(pp.points[0]);
          start = pp.points[0];
          break;
        case LineToAbsolute:
          poly.addHandle(pp.points[0]);
          end = pp.points[0];
          break;
        case ClosePath:
          poly.setClosedPolygone(true);
          poly.setHasArrowAtEnd(false);
          poly.setHasArrowAtFront(false);
          break;
      }
    }

    int lineStrength = SVGImporter.getLineStrength(n);
    Color lineColor = SVGImporter.getRGBColorForAttributeName("stroke", n);
    Color fillColor = SVGImporter.getRGBColorForAttributeName("fill", n);
    
    poly.setLineColor(lineColor);
    poly.setFillColor(fillColor);
    poly.setLineStrength(lineStrength);
    
    if(!poly.isClosedPolygone()) {
      for(int j = 0; j<2; j++) {
        n = null;
        for(; i<element.getChildNodes().getLength(); i++) {
          n = element.getChildNodes().item(i);
          if(n.getNodeName().equalsIgnoreCase("path")) {
            i++;
            break;
          }
        }
        if(n!=null)
          SVGImporter.setArrow(poly, start, end, n);
      }
    }
    poly.setMode(SVGElement.Mode.Normal);
    poly.deSelect();
    return poly;
  }
  
  private static void setArrow(SVGArrow elem, Point start, Point end, Node line) {
    if(line!=null && line.getNodeName().equalsIgnoreCase("path")) {
      Point mP = SVGImporter.
              getPathPointsForAttributeName("d", line).
              firstElement().points[0];
      SVGElement e = (SVGElement)elem;
      if (mP.distance(start) < mP.distance(end)) {
        start = mP;
        elem.setHasArrowAtEnd(true);
        e.getHandles().get(0).setLocation(mP);
      } else {
        end = mP;
        elem.setHasArrowAtFront(true);
        e.getHandles().get(e.getHandles().size()-1).setLocation(mP);
      }
    }
  }

  private static SVGElement importSVGRectangle(Element element)
          throws XPathExpressionException {

    SVGRectangle elem = new SVGRectangle();
    Node n = element;
    Node textNode = null;
    
    for(int i = 0; i<element.getChildNodes().getLength(); i++) {
      n = element.getChildNodes().item(i);
      if(n.getNodeName().equalsIgnoreCase("text"))
        textNode = n;
      if(n.getNodeName().equalsIgnoreCase("rect"))
        elem = SVGImporter.importRect(n);
    }
    
    if(textNode!=null) {
      elem.setMode(SVGElement.Mode.Edit);
      elem.setMode(SVGElement.Mode.TextEdit);
      SVGText text = SVGImporter.importText(textNode, elem.getBoundingBox(), false);
      elem.setText(text.getText());
      elem.setTextColor(text.getTextColor());
      elem.setTextAlignment(text.getTextAlignment());
      elem.setTextSize(text.getTextSize());
      elem.setTextOption(text.getTextOption());
      elem.setFont(text.getFont());
    }
    elem.setMode(SVGElement.Mode.Edit);
    elem.setMode(SVGElement.Mode.Normal);
    return elem;
   }

  
  private static SVGElement importSVGRoundSidedRectangle(Element element) {
    SVGRectangle rect = new SVGRectangle();
    
    Node n = null;
    Node rectNode = null;
    Node textNode = null;
    for(int i = 0; i<element.getChildNodes().getLength(); i++) {
      n = element.getChildNodes().item(i);
      if(n.getNodeName().equalsIgnoreCase("path"))
        rectNode = n;
      if(n.getNodeName().equalsIgnoreCase("text"))
        textNode = n;
    }
    if(rectNode==null)
      return rect;
    Vector<PathPoint> pathPoints = SVGImporter.getPathPointsForAttributeName("d", rectNode);
    for(PathPoint pp : pathPoints) {
      if(pp.type == PathPoint.Type.CurveToAbsolute) {
        if(rect.getHandles().get(0).getLocation().x > 0) {
          rect.getHandles().get(2).setLocation(pp.points[0]);
        rect.getHandles().get(1).setLocation(pp.points[0]);
        rect.getHandles().get(3).setLocation(pp.points[0]);
      } else
          rect.getHandles().get(0).setLocation(pp.points[0]);
        //rect.getHandles().get(1).setLocation(pp.points[0]);
        //rect.getHandles().get(3).setLocation(pp.points[0]);
        
      } else if(pp.type == PathPoint.Type.CurveToRelative) {
        
      }
    }
    
    int lineStrength = SVGImporter.getLineStrength(rectNode);
    Color lineColor = SVGImporter.getRGBColorForAttributeName("stroke", rectNode);
    Color fillColor = SVGImporter.getRGBColorForAttributeName("fill", rectNode);
    //rect.select();
    rect.setLineColor(lineColor);
    rect.setFillColor(fillColor);
    rect.setLineStrength(lineStrength);
    rect.setMode(SVGElement.Mode.Edit);
    rect.setMode(SVGElement.Mode.TextEdit);
    if(textNode!=null) {
      rect.setMode(SVGElement.Mode.Edit);
      rect.setMode(SVGElement.Mode.TextEdit);
      SVGText text = SVGImporter.importText(textNode, rect.getBoundingBox(), false);
      rect.setText(text.getText());
      rect.setTextColor(text.getTextColor());
      rect.setTextAlignment(text.getTextAlignment());
      rect.setTextSize(text.getTextSize());
      rect.setTextOption(text.getTextOption());
      rect.setFont(text.getFont());
    }
    rect.setType(SVGRectangle.Type.RoundSidedRectangle);
    rect.setMode(SVGElement.Mode.Edit);
    rect.setMode(SVGElement.Mode.Normal);
    
    return rect;
  }
  
  private static SVGElement importSVGEllipse(Element element)
          throws XPathExpressionException {
    SVGEllipse elem = new SVGEllipse();
    Node n = element;
    Node textNode = null;
    
    for(int i = 0; i<element.getChildNodes().getLength(); i++) {
      n = element.getChildNodes().item(i);
      if(n.getNodeName().equalsIgnoreCase("text"))
        textNode = n;
      if(n.getNodeName().equalsIgnoreCase("ellipse"))
        elem = SVGImporter.importEllipse(n);
    }

    if(textNode!=null) {
      elem.setMode(SVGElement.Mode.Edit);
      elem.setMode(SVGElement.Mode.TextEdit);
      SVGText text = SVGImporter.importText(textNode, elem.getBoundingBox(), false);
      elem.setText(text.getText());
      elem.setTextColor(text.getTextColor());
      elem.setTextAlignment(text.getTextAlignment());
      elem.setTextSize(text.getTextSize());
      elem.setTextOption(text.getTextOption());
      elem.setFont(text.getFont());
    }
    elem.setMode(SVGElement.Mode.Edit);
    elem.setMode(SVGElement.Mode.Normal);
    return elem;
  }
  
  private static SVGElement importSVGText(Element element)
          throws XPathExpressionException {

    SVGText elem = new SVGText();
    SVGRectangle rect = null;
    Node n = null;

    for(int i = 0; i<element.getChildNodes().getLength(); i++) {
      n = element.getChildNodes().item(i);
      if(n.getNodeName().equalsIgnoreCase("rect"))
        rect = SVGImporter.importRect(n);
      if(n.getNodeName().equalsIgnoreCase("text"))
        elem = SVGImporter.importText(n, rect.getBoundingBox(), true);
    }
    elem.setMode(SVGElement.Mode.Edit);
    if(rect!=null && elem!=null) {
      elem.setFillColor(rect.getFillColor());
      elem.setLineColor(rect.getLineColor());
      elem.setLineStrength(rect.getLineStrength());
      elem.getHandles().get(0).setLocation(rect.getHandles().get(0));
      elem.getHandles().get(1).setLocation(rect.getHandles().get(2));
    }
    elem.setMode(SVGElement.Mode.Normal);
 
    return elem;
  }

  private static SVGLine importLine(Node n) {
    if(n==null || !n.getNodeName().equalsIgnoreCase("line"))
     return null;

    Point start = SVGImporter.getPointForAttributeNames("x1", "y1", n);
    Point end = SVGImporter.getPointForAttributeNames("x2", "y2", n);
    Color lineColor = SVGImporter.getRGBColorForAttributeName("stroke", n);
    int lineStrength = SVGImporter.getLineStrength(n);
    
    SVGLine line = new SVGLine();
    line.getHandles().get(0).setLocation(start);
    line.getHandles().get(1).setLocation(end);
    line.setLineColor(lineColor);
    line.setLineStrength(lineStrength);
    line.setMode(SVGElement.Mode.Normal);
    return line;
  }
  
  private static SVGEllipse importEllipse(Node n) {
    if(n==null || !n.getNodeName().equalsIgnoreCase("ellipse"))
     return null;
    
    SVGEllipse elem = new SVGEllipse();

    Point center = SVGImporter.getPointForAttributeNames("cx", "cy", n);
    Point radius = SVGImporter.getPointForAttributeNames("rx", "ry", n);
    int lineStrength = SVGImporter.getLineStrength(n);
    Color lineColor = SVGImporter.getRGBColorForAttributeName("stroke", n);
    Color fillColor = SVGImporter.getRGBColorForAttributeName("fill", n);
    Vector<Point> pts = new Vector<Point>();
    pts.add(new Point(center.x - radius.x, center.y - radius.y));
    pts.add(new Point(center.x + radius.x, center.y + radius.y));
    elem.getHandles().set(0, pts.get(0));
    elem.getHandles().set(1, pts.get(0));
    elem.getHandles().set(3, pts.get(0));
    elem.getHandles().set(2, pts.get(1));
    //elem.select();
    elem.setLineColor(lineColor);
    elem.setFillColor(fillColor);

    elem.setLineStrength(lineStrength);
    elem.setMode(SVGElement.Mode.Normal);
    return elem;
  }
  
  private static SVGRectangle importRect(Node n) {
    if(n==null || !n.getNodeName().equalsIgnoreCase("rect"))
     return null;
    
    SVGRectangle elem = new SVGRectangle();
    Point start = SVGImporter.getPointForAttributeNames("x", "y", n);
    int w = SVGImporter.getWidth(n);
    int h = SVGImporter.getHeight(n);
    int lineStrength = SVGImporter.getLineStrength(n);
    Color lineColor = SVGImporter.getRGBColorForAttributeName("stroke", n);
    Color fillColor = SVGImporter.getRGBColorForAttributeName("fill", n);
    if(SVGImporter.isRoundedRect(n))
      elem.setType(SVGRectangle.Type.RoundedRectangle);
    Vector<Point> pts = new Vector<Point>();
    pts.add(start);
    pts.add(new Point(start.x + w, start.y + h));
    elem.getHandles().set(0, start);
    elem.getHandles().set(1, start);
    elem.getHandles().set(3, start);
    elem.getHandles().set(2, new Point(start.x + w, start.y + h));
    //elem.select();
    elem.setLineColor(lineColor);
    elem.setFillColor(fillColor);

    elem.setLineStrength(lineStrength);
    elem.setMode(SVGElement.Mode.Normal);

    return elem;
  }
  
  private static SVGText importText(Node n, Rectangle boundingRect, boolean fillEmptyLines) {
    if(n==null || !n.getNodeName().equalsIgnoreCase("text"))
     return null;
    
    SVGText elem = new SVGText();
    Point textPoint = SVGImporter.getPointForAttributeNames("x", "y", n);
    Point start = textPoint;
    Point end = start;
    if(boundingRect != null) {
      start = boundingRect.getLocation();
      end = new Point(start.x + boundingRect.width, start.y + boundingRect.height);
    }
    
    String font = SVGImporter.getStringValueForAttributeName("font-family", n);
    int textSize = SVGImporter.getIntValueForAttributeName("font-size", n);
    String alignment = SVGImporter.getStringValueForAttributeName("text-anchor", n);
    Color fontColor = SVGImporter.getRGBColorForAttributeName("fill", n);
    // text-decoration = "line-through,underline"
    // font-style="italic" font-weight="bold"
    // Checkt and Set strikethrough & underlined
    Node decoration = n.getAttributes().getNamedItem("text-decoration");
    if(decoration!=null) {
      if(decoration.getNodeValue().contains("line-through"))
        elem.addTextOption(OptionManager.TextOption.strikethrough);
      if(decoration.getNodeValue().contains("underline"))
        elem.addTextOption(OptionManager.TextOption.underlined);
    }
    
    Node style = n.getAttributes().getNamedItem("font-style");
    if(style!=null && style.getNodeValue().contains("italic"))
      elem.addTextOption(OptionManager.TextOption.italic);
    Node weight = n.getAttributes().getNamedItem("font-weight");
    if(weight!=null && weight.getNodeValue().contains("bold"))
      elem.addTextOption(OptionManager.TextOption.bold);
    Node tmp = null;
    StringBuilder text = new StringBuilder();
    int height = 0;
    int dy = 0;
    int offset = 0;
    for(int i = 0; i<n.getChildNodes().getLength(); i++) {
      tmp = n.getChildNodes().item(i);
      if(tmp.getNodeName().equalsIgnoreCase("tspan") && tmp.hasChildNodes() && tmp.getFirstChild().getNodeType() == Node.TEXT_NODE) {
        dy = SVGImporter.getIntValueForAttributeName("dy", tmp);
        height+=dy;
        offset = (dy/textSize);
        for(int j = 1; j<offset; j++)
          text.append("\n");
        text.append(tmp.getFirstChild().getNodeValue());
        text.append("\n");
      }
      if(i==n.getChildNodes().getLength()-1 && boundingRect != null && fillEmptyLines) {
        int rest = boundingRect.height-height;
        offset = (rest/textSize);
        for(int j = 1; j<offset-1; j++)
          text.append("\n");
      }
    }
    if(text.length()>0)
      text.deleteCharAt(text.length()-1);

    elem.getHandles().set(0, start);
    elem.getHandles().set(1, end);
    elem.setFont(Font.decode(font));
    elem.setTextSize(textSize);
    elem.setTextColor(fontColor);
    elem.setText(text.toString());
    switch(alignment) {
      case "start":
        elem.setTextAlignment(OptionManager.Alignment.left);
        break;
      case "middle":
        elem.setTextAlignment(OptionManager.Alignment.center);
        break;
      case "end":
        elem.setTextAlignment(OptionManager.Alignment.right);
        break;
    }
    elem.setMode(SVGElement.Mode.Edit);
    elem.setMode(SVGElement.Mode.TextEdit);
    elem.setMode(SVGElement.Mode.Edit);
    elem.setMode(SVGElement.Mode.Normal);
    return elem;
  }
  
  private void printSVGElement(Element eElement)
          throws XPathExpressionException {
    XPath xpath = XPathFactory.newInstance().newXPath();
    NodeList nodeList = (NodeList) xpath.evaluate("*", eElement,
                                                  XPathConstants.NODESET);
    for (int i = 0; i < nodeList.getLength(); i++) {
      NamedNodeMap attributes = nodeList.item(i).getAttributes();
    }
  }

  private static boolean isRoundedRect(Node rect) {
    if(rect.getAttributes().getNamedItem("rx")!=null)
      return true;
    return false;
  }
    
  private static Point getPointForAttributeNames(String xAttribute, String yAttribute, Node node) {
    String x = node.getAttributes().getNamedItem(xAttribute.toLowerCase()).
            getNodeValue();
    String y = node.getAttributes().getNamedItem(yAttribute.toLowerCase()).
            getNodeValue();

    Point p = new Point(Integer.parseInt(x), Integer.parseInt(y));
    return p;
  }

  private static int getWidth(Node node) {
    String width = node.getAttributes().getNamedItem("width").getNodeValue();
    return Integer.parseInt(width);
  }

  private static int getHeight(Node node) {
    String height = node.getAttributes().getNamedItem("height").getNodeValue();
    return Integer.parseInt(height);
  }

  private static int getLineStrength(Node node) {
    if (node.getAttributes().getNamedItem("stroke-width").getNodeValue().contains(".")) {
      double stroke = Double.parseDouble(node.getAttributes().getNamedItem("stroke-width").getNodeValue());
      if(stroke==0.25) // Hairline
        return 0;
      if(stroke == 0.0) // No Line
        return -1;
      return (int)Math.round(stroke);
    } else
      return Integer.parseInt(node.getAttributes().getNamedItem(
              "stroke-width").getNodeValue());
  }

  private static Color getRGBColorForAttributeName(String attribute, Node node) {
    String color = node.getAttributes().getNamedItem(attribute).getNodeValue();
    if (color.equalsIgnoreCase("none"))
      return new Color(255, 255, 255, 0);
    color = color.substring(4);
    color = color.substring(0, color.length() - 1);
    return new Color(Integer.parseInt(color.split(",")[0]),
                     Integer.parseInt(color.split(",")[1]),
                     Integer.parseInt(color.split(",")[2]));
  }

  private static int getIntValueForAttributeName(String attribute, Node node) {
    return Integer.parseInt(node.getAttributes().getNamedItem(attribute).
            getNodeValue());
  }

  private static String getStringValueForAttributeName(String attribute, Node node) {
    return node.getAttributes().getNamedItem(attribute).getNodeValue();
  }
  
  private static int getMinIndex(int[] indizes) {
    int idx = Integer.MAX_VALUE;
    for(int i = 0; i<indizes.length; i++)
      if(i<idx)
        idx = i;
    return idx;
  }
  
  private static Vector<PathPoint> getPathPointsForAttributeName(String attribute, Node node) {
    Vector<PathPoint> points = new Vector<PathPoint>();
    String split = " ";
    String pathString = node.getAttributes().getNamedItem(attribute).
            getNodeValue().trim();
    StringTokenizer tokenizer = new StringTokenizer(pathString, "MmLlCcSsQqTtAaZz");
    int tokIdx = 0;
    while(tokenizer.hasMoreTokens()) {
      String tok = tokenizer.nextToken().trim();
      tokIdx = pathString.indexOf(tok, tokIdx)-1;
      char c = pathString.charAt(tokIdx);
      String[] pointTokens = tok.split(" ");
      Point[] p = new Point[pointTokens.length];
      for(int i = 0; i<pointTokens.length; i++) {
        if(pointTokens[i].contains(".")) {
          p[i] = new Point((int)Double.parseDouble(pointTokens[i].split(",")[0].trim()), 
                           (int)Double.parseDouble(pointTokens[i].split(",")[1].trim()));
        } else {
          p[i] = new Point(Integer.parseInt(pointTokens[i].split(",")[0].trim()), 
                           Integer.parseInt(pointTokens[i].split(",")[1].trim()));
        }
      }

      PathPoint pp = null;
      switch(c) {
        case 'M':
          pp = new PathPoint(PathPoint.Type.MoveToAbsolute, p);
          break;
        case 'm':
          pp = new PathPoint(PathPoint.Type.MoveToRelative, p);
          break;
        case 'L':
          pp = new PathPoint(PathPoint.Type.LineToAbsolute, p);
          break;
        case 'l':
          pp = new PathPoint(PathPoint.Type.LineToRelative, p);
          break;
        case 'C':
          pp = new PathPoint(PathPoint.Type.CurveToAbsolute, p);
          break;
        case 'c':
          pp = new PathPoint(PathPoint.Type.CurveToRelative, p);
          break;
        case 'S':
          pp = new PathPoint(PathPoint.Type.SmoothCurveToAbsolute, p);
          break;
        case 's':
          pp = new PathPoint(PathPoint.Type.SmoothCurveToRelative, p);
          break;
        case 'Q':
          pp = new PathPoint(PathPoint.Type.QuadCurveToAbsolute, p);
          break;
        case 'q':
          pp = new PathPoint(PathPoint.Type.QuadCurveToRelative, p);
          break;
        case 'T':
          pp = new PathPoint(PathPoint.Type.SmoothQuadCurveToAbsolute, p);
          break;
        case 't':
          pp = new PathPoint(PathPoint.Type.SmoothQuadCurveToRelative, p);
          break;
        case 'A':
          break;
        case 'a':
          break;
        case 'Z':
        case 'z':
          pp = new PathPoint(PathPoint.Type.ClosePath, null);
          points.add(pp);
          return points;
      }
      if(pp!=null)
        points.add(pp);
    }
    if(pathString.endsWith("z") || pathString.endsWith("Z"))
      points.add(new PathPoint(PathPoint.Type.ClosePath, null));
    return points;
  }
  
  private static class PathPoint {
    public enum Type {MoveToAbsolute, MoveToRelative, 
                      LineToAbsolute, LineToRelative, 
                      CurveToAbsolute, CurveToRelative, 
                      SmoothCurveToAbsolute, SmoothCurveToRelative, 
                      QuadCurveToAbsolute, QuadCurveToRelative, 
                      SmoothQuadCurveToAbsolute, SmoothQuadCurveToRelative, 
                      ArcCurveToAbsolute, ArcCurveToRelative, 
                      ClosePath
    };
    
    public Type type = null;
    public Point[] points = null;
    public PathPoint(Type t, Point[] pts) {
      this.type = t;
      this.points = pts;
    }
  }
}
