 /**----------------------------------------------------
 * File:      Canvas.java
 * Package:   at.hoertler.svgeditor.display.controls
 * Project:   SVGEditor
 * ----------------------------------------------------
 * @author Philipp Hörtler BSc
 * @version 1.0.0
 * Date: 2013-03-24
 * ----------------------------------------------------
 * Description:
 *
 * The canvas receives mouse actions.
 * It adapts the mousepositions to the magnetic grid and forwards 
 * the events to the correct receivers.
 */
package at.hoertler.svgeditor.display.controls;

import at.hoertler.svgeditor.display.toolbar.ToolBarSelectionButton;
import at.hoertler.svgeditor.display.toolbar.ToolBarDrawingButton;
import at.hoertler.svgeditor.Editor;
import at.hoertler.svgeditor.common.HelperFunction;
import at.hoertler.svgeditor.common.KeyInputManager;
import at.hoertler.svgeditor.display.toolbar.ToolBarButton;
import at.hoertler.svgeditor.display.toolbar.ToolBarManager;
import at.hoertler.svgeditor.event.ActionPerformedEvent;
import at.hoertler.svgeditor.event.ActionPerformedListener;
import at.hoertler.svgeditor.event.CanvasStateEvent;
import at.hoertler.svgeditor.event.CanvasStateListener;
import at.hoertler.svgeditor.event.OptionChangedEvent;
import at.hoertler.svgeditor.event.OptionChangedListener;
import at.hoertler.svgeditor.options.OptionManager;
import at.hoertler.svgeditor.svg.elements.SVGElement;
import at.hoertler.svgeditor.svg.elements.SVGSelection;
import at.hoertler.svgeditor.svg.elements.SVGText;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.geom.Point2D;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Vector;
import javax.swing.JPanel;


public class Canvas
        extends JPanel
        implements OptionChangedListener,
                   ActionPerformedListener,
                   MouseListener,
                   MouseMotionListener,
                   ComponentListener {

  private Editor editor = null;
  private ToolBarManager toolBarManager = null;
  private SVGSelection selection = null;
  private SVGElement activeElement = null;
  private int zoomFactor = 1;
  private StringBuilder inputText = new StringBuilder("");
 private Point startPoint = null;
  private Point correction = null;
  private MagneticGrid magneticGrid = null;
  private Point grabPoint = null, releasePoint = null;
  private int topMargin = 10;
  private int leftMargin = 10;
  private LinkedList<CanvasStateListener> listeners = null;
  private boolean dragged = false;

  private class Statistik {
    int cnt = 0;
    double time;

    public void add(double t) {
      ++cnt;
      time+=t;
    }
  }
  
  private HashMap<Integer,Statistik> stats = new HashMap<Integer,Statistik>();

  /**
   * @return the zoomFactor
   */
  public int getZoomFactor() {
    return zoomFactor;
  }

  /**
   * @param zoomFactor the zoomFactor to set
   */
  public void setZoomFactor(int zoomFactor) {
    this.zoomFactor = zoomFactor;
  }

  public Canvas(Editor aEditor, ToolBarManager tbm, OptionManager om) {
    this.editor = aEditor;
    this.toolBarManager = tbm;
    this.setDoubleBuffered(true); // display efficiency
    this.magneticGrid = new MagneticGrid(800, 600);
    this.setBackground(Color.WHITE);
    this.addMouseListener(this);
    this.addMouseMotionListener(this);
    om.addOptionChangedListener(this);
    this.adjustCanvasSize();
    this.listeners = new LinkedList<CanvasStateListener>();
    this.addComponentListener(this);
    KeyInputManager.registerCanvas(this);
  }
  
    /**
   * Adds an listener to the Canvas.
   *
   * @param l CanvasStateListener
   */
  public void addCanvasStateListener(CanvasStateListener l) {
    listeners.add(l);
  }

  /**
   * Removes an listener from the Canvas.
   *
   * @param l CanvasStateListener
   */
  public void removeCanvasStateListener(CanvasStateListener l) {
    listeners.remove(l);
  }

  /**
   * Fires the given CanvasStateEvent to all listeners.
   *
   * @param e CanvasStateListener that the listeners should recognize.
   */
  public void fireCanvasStateEvent(CanvasStateEvent e) {
    for (CanvasStateListener l : this.listeners) {
      l.canvasStateChanged(e);
    }
  }

  @Override
  public void mouseClicked(MouseEvent e) {
    this.handleZoom(e);
    if (e.getButton() == MouseEvent.BUTTON1) {
      if (!toolBarManager.getActiveButton().getClass().
              isAssignableFrom(ToolBarSelectionButton.class)) {
        
        this.selection = null;
        if(this.activeElement == null || ((!this.activeElement.isExtendedCreateMode() && !this.activeElement.isExtendedEditMode())
                && (!this.activeElement.getClass().isAssignableFrom(SVGText.class) || !this.activeElement.isTextEditMode()))) {
          
          this.toolBarManager.switchToSelectionButton();
          this.activeElement = null;
          this.editor.getElementManager().deselectAll();
        } else {
          this.handleGrid(e);
          this.activeElement.handleMouseClicked(e);
        }
      } else { 
        if(this.selection != null)
          this.selection.handleMouseClicked(e);
      }
    }
    this.selection=null;
  }

  @Override
  public void mousePressed(MouseEvent e) {
    ToolBarButton tbb = this.toolBarManager.getActiveButton();
    this.handleZoom(e);
    if (e.getButton() == MouseEvent.BUTTON1) {
      if (tbb.getClass().isAssignableFrom(ToolBarSelectionButton.class)) {
        this.activeElement = null;
        this.selection = null;
        this.selection = ((ToolBarSelectionButton) tbb).getSelection();
        this.selection.handleMousePressed(e);
      } else if (tbb.getClass().isAssignableFrom(ToolBarDrawingButton.class)) {
        this.handleGrid(e);
        this.selection = null;
        if(this.activeElement == null || (this.activeElement.getClass().isAssignableFrom(SVGText.class) && this.activeElement.isTextEditMode() && !this.activeElement.getText().isEmpty())) {
          this.activeElement = null;
          this.editor.getElementManager().deselectAll();
          this.activeElement = ((ToolBarDrawingButton) tbb).getElement();
        }
        if(this.activeElement.isInitialCreateMode())
          this.activeElement.handleMousePressed(e);
      }
      this.grabPoint = e.getPoint();
      this.releasePoint = e.getPoint();
      //this.fireElementManagerStateEvent(new CanvasStateEvent(this.grabPoint, new Point(0,0), e, this.editor.getElementManager().getSelectedElements()));
    }
  }

  @Override
  public void mouseDragged(MouseEvent e) {
    this.dragged = true;
    this.handleZoom(e);
    Point deltaPoint = new Point(0,0);
    if(this.grabPoint!=null && this.releasePoint!=null) {
      deltaPoint = HelperFunction.
                  calculateTranslationPoint(this.grabPoint,e.getPoint());

      this.releasePoint = e.getPoint();
      if(!e.isAltDown() && (this.selection==null || this.selection.getActiveElement()!=null)) { // if ALT is pressed -> ignore Grid
        Point trans = this.getGridTranslationPoint(deltaPoint);
        deltaPoint.translate(trans.x, trans.y);
      }
      
      deltaPoint.translate(this.grabPoint.x, this.grabPoint.y);
      deltaPoint = HelperFunction.calculateTranslationPoint(e.getPoint(),deltaPoint);
      if(e.isShiftDown()) { // if SHIFT is pressed -> fix X- or Y-Axis
        Point trans = HelperFunction.calculateTranslationPoint(e.getPoint(), this.grabPoint);
        if(Math.abs(e.getPoint().x-this.grabPoint.x) <= Math.abs(e.getPoint().y-this.grabPoint.y))
          deltaPoint.x = trans.x;
        else
          deltaPoint.y = trans.y;
      }
    }
    e.translatePoint(deltaPoint.x, deltaPoint.y);
    
    if (this.selection != null) {
      this.selection.handleMouseDragged(e);
    } else if (this.activeElement != null && this.activeElement.isInitialCreateMode()) {
      if(e.isControlDown()) {
        
        Point trans = HelperFunction.calculateTranslationPoint(this.grabPoint, e.getPoint());
        if(Math.abs(e.getPoint().x-this.grabPoint.x) <= Math.abs(e.getPoint().y-this.grabPoint.y)) {
          deltaPoint.x = trans.y-trans.x;
          deltaPoint.y = 0;
        } else {
          deltaPoint.x = 0;
          deltaPoint.y = trans.x-trans.y;
        }
        e.translatePoint(deltaPoint.x, deltaPoint.y);
        this.activeElement.handleMouseDragged(e);
      } else {
        this.activeElement.handleMouseDragged(e);
      }
    }
    this.fireCanvasStateEvent(new CanvasStateEvent(this.grabPoint, e.getPoint(), null, this.editor.getElementManager().getSelectedElements())); 
    // Only repaint if really neccessary
    tryRepaint(e.getPoint());
  }

  @Override
  public void mouseReleased(MouseEvent e) {
    this.dragged = false;
    this.handleZoom(e);
    if (e.getButton() == MouseEvent.BUTTON1) {
      if (this.selection != null) {
        this.selection.handleMouseReleased(e);
        this.adjustCanvasSize(this.selection.getActiveElement());
        this.fireCanvasStateEvent(new CanvasStateEvent(
                this.grabPoint, 
                null, 
                null, 
                this.editor.getElementManager().getSelectedElements()
                ));
      } else if (this.activeElement != null) {
        this.handleGrid(e);
        this.activeElement.handleMouseReleased(e);
        if(!this.editor.getElementManager().getAllElements().contains(this.activeElement)) {
          boolean added = false;
          if (this.activeElement.validate()) {
            added = this.editor.getElementManager().addElement(this.activeElement);
            this.adjustCanvasSize(this.activeElement);
          } 
          if (!added) this.toolBarManager.switchToSelectionButton();
        }

        if(!this.activeElement.isExtendedCreateMode() && 
           (!this.activeElement.getClass().isAssignableFrom(SVGText.class) || 
            !this.activeElement.isTextEditMode())
            ) {
          this.activeElement = null;
        }

      }
    }
    this.grabPoint = new Point(0,0);
    this.repaint();
  }

  @Override
  public void mouseMoved(MouseEvent e) {
    this.handleZoom(e);

    ToolBarButton tbb = this.toolBarManager.getActiveButton();
    if (tbb.getClass().isAssignableFrom(ToolBarSelectionButton.class)) {
      SVGElement elem = this.editor.getElementManager().getHitElement(e.getPoint());
      for (SVGElement tmp : this.editor.getElementManager().getSelectedElements()) {
        if (tmp.hitHandle(e.getPoint())) {
          elem = tmp;
          break;
        }
      }

      if (elem != null) {
        Cursor c = null;
        if (elem.hitHandle(e.getPoint())) {
          c = elem.getHandleCursor();
        } else if (elem.isSelected()) {
          switch(elem.getMode()) {
            case ExtendedEdit:
              c = elem.getExtendedEditCursor();
              break;
            case ExtendedCreate:
              c = elem.getExtendedCreateCursor();
              break;
            case TextEdit:
              c = elem.getTextEditCursor();
              break;
            default:
              c = elem.getEditCursor();
              break;
          }
        } else {
          c = elem.getNormalCursor();
        }
        setCursor(c);
      } else if (this.getCursor().getType() != Cursor.DEFAULT_CURSOR) {
        Cursor c = new Cursor(Cursor.DEFAULT_CURSOR);
        setCursor(c);
      }
    } else if(tbb.getClass().isAssignableFrom(ToolBarDrawingButton.class)) {
      setCursor((((ToolBarDrawingButton) tbb).getElement()).getInitialCreateCursor());
      if(this.activeElement != null && this.activeElement.isExtendedCreateMode()) {
        this.handleGrid(e);
        this.activeElement.handleMouseMoved(e);
        this.tryRepaint(e.getPoint());
      }
    }
    this.fireCanvasStateEvent(new CanvasStateEvent(
            null, 
            null, 
            e, 
            this.editor.getElementManager().getSelectedElements()
            ));
  }

  @Override
  public void mouseEntered(MouseEvent e) {
    //this.handleZoom(e);
    //throw new UnsupportedOperationException("Not supported yet.");
  }

  @Override
  public void mouseExited(MouseEvent e) {
    //this.handleZoom(e);
  }

  @Override
  public void paint(Graphics g) {
    long start = System.currentTimeMillis();
    //System.out.println("\n---------------------------------------------");
    
    super.paint(g);
    //System.out.println(new Date() + "\t\t- Repaint SU.t="+(System.currentTimeMillis()-start));
    this.magneticGrid.paint((Graphics2D)g.create());
    //System.out.println(new Date() + "\t\t- Repaint Grid.t="+(System.currentTimeMillis()-start));
    // Paint elements
    Vector<SVGElement> handles = new Vector<SVGElement>();
    Vector<SVGElement> remove = new Vector<SVGElement>();
    for (SVGElement elem : this.editor.getElementManager().getAllElements()) {
      if(!elem.isSelected() && !elem.validate())
        remove.add(elem);
      else {
        elem.paint(g, this.getZoomFactor());
        if (elem.isSelected()) // Store handles if element is selected
          handles.add(elem);
      }
    }
    //System.out.println(new Date() + "\tElems: \t" + this.editor.getElementManager().getAllElements().size() + 
    //        "\t- Repaint ELEM.t\t"+(System.currentTimeMillis()-start));
    int size = this.editor.getElementManager().getAllElements().size();
    long time = (System.currentTimeMillis()-start);
    if(!stats.containsKey(size))
      stats.put(size, new Statistik());
    stats.get(size).add(time);

    // Remove elements that dont validate
    if(!remove.isEmpty())
      this.editor.getElementManager().getAllElements().removeAll(remove);
    
    // Paint selection or element while creating
    if (this.selection != null) {
      this.selection.paint(g, this.getZoomFactor());
    } else if (this.activeElement != null) {
      if(!this.editor.getElementManager().getAllElements().contains(this.activeElement))
        this.activeElement.paint(g, this.getZoomFactor());
    }
    //System.out.println(new Date() + "\t\t- Repaint SEL/ACT.t="+(System.currentTimeMillis()-start));
    // Paint Handles of selected elements
    for (SVGElement elem : handles) 
      elem.paintHandles(g, this.getZoomFactor());
    //System.out.println(new Date() + "\t\t- Repaint HANDLES.t="+(System.currentTimeMillis()-start));
  }

  private boolean tryRepaint(Point p) {
    boolean repaint = false;
    if (this.startPoint == null) {
      this.startPoint = p;
    } else {
      if(startPoint.distance((Point2D)p) >= this.magneticGrid.getMagneticGapSize()) {
      //if (!startPoint.equals(p)) {
        repaint = true;
        this.startPoint = p;
      }
    }
    long start = System.currentTimeMillis();
    if (repaint) {
      //System.out.println("\n---------------------------------------------");
      //System.gc();
      //System.out.println(new Date() + "\t- Try Repaint GC.t="+(System.currentTimeMillis()-start));
      //System.runFinalization();
      //System.out.println(new Date() + "\t- Try Repaint Fin.t="+(System.currentTimeMillis()-start));
      this.repaint();
      //System.out.println(new Date() + "\t- Try Repaint Rep.t="+(System.currentTimeMillis()-start));
    }
    return repaint;
  }

  /**
   * Method handleGrid.
   *
   * @param e MouseEvent
   */
  private void handleGrid(MouseEvent e) {
    if(e.isAltDown())
      return;
    if (e.getPoint().x % this.magneticGrid.getMagneticGapSize() != 0 || e.getPoint().y % this.magneticGrid.getMagneticGapSize() != 0) {
      Point p = this.getGridTranslationPoint(e.getPoint());
      e.translatePoint(p.x, p.y);
    }
  }

  /**
   * Method getGridTranslationPoint.
   *
   * @param p Point
   *
   * @return Point
   */
  private Point getGridTranslationPoint(Point p) {
    Point ret = new Point();
    int x = p.x;
    int y = p.y;
    if (Math.abs(x % this.magneticGrid.getMagneticGapSize()) > ((double) this.magneticGrid.getMagneticGapSize() / 2)) {
      ret.x = (this.magneticGrid.getMagneticGapSize() - Math.abs(x % this.magneticGrid.getMagneticGapSize()))*(x/Math.abs(x));
    } else {
      ret.x = -(x % this.magneticGrid.getMagneticGapSize());
    }

    if (Math.abs(y % this.magneticGrid.getMagneticGapSize()) > ((double) this.magneticGrid.getMagneticGapSize() / 2)) {
      ret.y = (this.magneticGrid.getMagneticGapSize() - Math.abs(y % this.magneticGrid.getMagneticGapSize()))*(y/Math.abs(y));
    } else {
      ret.y = -(y % this.magneticGrid.getMagneticGapSize());
    }
    return ret;
  }

  /**
   * Method handleZoom.
   *
   * @param e MouseEvent
   */
  private void handleZoom(MouseEvent e) {
    if (zoomFactor > 1) {
      Point trans = new Point(0, 0);
      trans.x = e.getPoint().x / this.getZoomFactor() - e.getPoint().x;
      trans.y = e.getPoint().y / this.getZoomFactor() - e.getPoint().y;
      e.translatePoint(trans.x, trans.y);
    }
  }

  @Override
  public void optionChanged(OptionChangedEvent e) {
    switch (e.getOption()) {
      case gridWideness:
        this.magneticGrid.setMagneticGapSize((int) e.getValue());
        break;
      case zoomFactor:
        this.magneticGrid.setZoom((int) e.getValue());
        this.setZoomFactor((int) e.getValue());
        this.adjustCanvasSize();
        break;
      default:
        break;
    }
    this.repaint();
  }

  private void adjustCanvasSize() {
    Rectangle r = this.getBounds();
    if(this.getParent()!=null)
      r = this.getParent().getBounds();
    r.setLocation(0,0);
    this.setBounds(r);
    this.setPreferredSize(new Dimension(r.width,r.height));
    for (SVGElement elem : this.editor.getElementManager().getAllElements()) {
      this.adjustCanvasSize(elem);
    }
    int i = 0;
  }

  private void adjustCanvasSize(SVGElement elem) {
    if (elem == null) {
      return;
    }

    if (elem.getBoundingBox().x < this.leftMargin || elem.getBoundingBox().y < this.topMargin) {

      Point trans = new Point(0, 0);
      if (elem.getBoundingBox().x < this.leftMargin) {
        trans.x = this.leftMargin - elem.getBoundingBox().x;
      }
      if (elem.getBoundingBox().y < this.topMargin) {
        trans.y = this.topMargin - elem.getBoundingBox().y;
      }
      for (SVGElement e : this.editor.getElementManager().getAllElements()) {
        e.translate(trans);
      }
      this.adjustCanvasSize();
      return;
    }

    if (this.getBounds().getHeight() < (elem.getBoundingBox().y * this.zoomFactor + elem.
                                        getBoundingBox().getHeight() * this.zoomFactor + this.topMargin)) {
      Dimension d = new Dimension((int) this.getBounds().getWidth(),
                                  (int) (elem.getBoundingBox().y * this.zoomFactor + elem.
                                         getBoundingBox().getHeight() * this.zoomFactor + this.topMargin));
      this.setPreferredSize(d);
      this.setMinimumSize(d);
      this.setBounds(this.getBounds().x, this.getBounds().y, (int) d.getWidth(),
                     (int) d.getHeight());
      this.setMagneticGridSize();
    }
    
    if (this.getBounds().getWidth() < (elem.getBoundingBox().x * this.zoomFactor + elem.
                                       getBoundingBox().getWidth() * this.zoomFactor + this.leftMargin)) {
      Dimension d = new Dimension(
              (int) (elem.getBoundingBox().x * this.zoomFactor + elem.
                     getBoundingBox().getWidth() * this.zoomFactor + this.leftMargin),
              (int) this.getBounds().getHeight());
      this.setPreferredSize(d);
      this.setMinimumSize(d);
      this.setBounds(this.getBounds().x, this.getBounds().y, (int) d.getWidth(),
                     (int) d.getHeight());
      this.setMagneticGridSize();
    }
    
  }

  @Override
  public void actionPerformed(ActionPerformedEvent e) {
    switch (e.getAction()) {
      case showMagneticGrid:
        this.magneticGrid.setMagneticGridDisplayed((boolean)e.getValue());
        this.repaint();
        System.out.println("Elems\t#\tTime");
        for ( Map.Entry<Integer,Statistik> elem : stats.entrySet())
          System.out.println(elem.getKey() + "\t" + elem.getValue().cnt + "\t" + elem.getValue().time + "\t" + elem.getValue().time/elem.getValue().cnt);
        break;
      case switchToSelection:
        this.toolBarManager.switchToSelectionButton();
        break;
      default:
        this.repaint();
    }
  }

  private void setMagneticGridSize() {
    if(this.getParent().getBounds().contains(this.getBounds())) {
      magneticGrid.setHeight((int)this.getParent().getBounds().getHeight());
      magneticGrid.setWidth((int)this.getParent().getBounds().getWidth());
    } else {
      magneticGrid.setHeight((int)this.getBounds().getHeight());
      magneticGrid.setWidth((int)this.getBounds().getWidth());
    }
System.out.println("MagneticGrid");
  }

  @Override
  public void componentResized(ComponentEvent e) {
    this.setMagneticGridSize();
  }

  @Override
  public void componentMoved(ComponentEvent e) { }

  @Override
  public void componentShown(ComponentEvent e) { }

  @Override
  public void componentHidden(ComponentEvent e) { }

  public void updateMouse(KeyEvent e) {
    if(this.getMousePosition() == null) return;
    
    int modifier = MouseEvent.BUTTON1_MASK;
    if(e!=null)
      modifier = e.getModifiers();
    if(this.dragged)
      this.mouseDragged(new MouseEvent(this, MouseEvent.MOUSE_DRAGGED, System.nanoTime(), modifier, this.getMousePosition().x, this.getMousePosition().y, 0, false));
    else
      this.mouseMoved(new MouseEvent(this, MouseEvent.MOUSE_MOVED, System.nanoTime(), 0, this.getMousePosition().x, this.getMousePosition().y, 0, false));
  }

}
