 /**----------------------------------------------------
 * File:      MenuBarManager.java
 * Package:   at.hoertler.svgeditor.display.menubar
 * Project:   SVGEditor
 * ----------------------------------------------------
 * @author Philipp Hörtler BSc
 * @version 1.0.0
 * Date: 2013-03-24
 * ----------------------------------------------------
 * Description:
 *
 * The MenuBarManager builds up the whole menubar.
 * The Menubar is used to fire Actions and to display some state informations.
 */
package at.hoertler.svgeditor.display.menubar;

import at.hoertler.svgeditor.Editor;
import at.hoertler.svgeditor.Reader;
import at.hoertler.svgeditor.action.ActionManager;
import at.hoertler.svgeditor.event.ActionPerformedEvent;
import at.hoertler.svgeditor.event.CanvasStateEvent;
import at.hoertler.svgeditor.event.CanvasStateListener;
import at.hoertler.svgeditor.event.ElementManagerStateEvent;
import at.hoertler.svgeditor.event.ElementManagerStateListener;
import java.awt.BasicStroke;
import java.awt.Component;
import java.awt.ComponentOrientation;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.geom.Point2D;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.KeyStroke;
import javax.swing.SwingConstants;


public class MenuBarManager
        implements ElementManagerStateListener, CanvasStateListener {

  private Editor editor = null;
  private JMenuBar menuBar = null;
  private JMenuItem newImage, openImage, lastOpenedImages, saveImage, 
                    saveImageAs, insertImage, printImage, exitApplication;
  private JMenuItem undo, redo, delete, cut, copy, paste, duplicate, 
                    takePicture, group, ungroup, selectAll, invertSelection, 
                    deselect, bringToFront, sendToBack, showMagGrid, hideMagGrid;
  private JMenu mousePosition, delta, alignment, magGrid;
  private JButton leaveGroup;

  public MenuBarManager(Editor editor) {
    this.editor = editor;
    this.menuBar = new JMenuBar();

    this.menuBar.add(createFileMenu());
    this.menuBar.add(createEditMenu());

    this.mousePosition = new JMenu("");
    this.mousePosition.setEnabled(false);
    this.mousePosition.setPreferredSize(new Dimension(60,this.mousePosition.getHeight()));
    this.delta = new JMenu("");
    this.delta.setEnabled(false);
    this.delta.setPreferredSize(new Dimension(80,this.mousePosition.getHeight()));
    this.delta.setIcon(new Icon() {

      @Override
      public void paintIcon(Component c, Graphics g, int x, int y) {
        Graphics2D g2 = (Graphics2D) g;
        g2.setColor(c.getForeground());
        g2.setStroke(new BasicStroke(1, BasicStroke.CAP_BUTT,
                                     BasicStroke.JOIN_MITER));
        g2.drawLine(x + 8, y + 2, x + 14, y + 14);
        g2.drawLine(x + 14, y + 14, x + 2, y + 14);
        g2.drawLine(x + 2, y + 14, x + 8, y + 2);
      }

      @Override
      public int getIconWidth() {
        return 16;
      }

      @Override
      public int getIconHeight() {
        return 16;
      }
    });
    this.delta.setDisabledIcon(this.delta.getIcon());
    this.menuBar.add(mousePosition);
    this.menuBar.add(delta);
  }

  /**
   * @return the editor
   */
  public Editor getEditor() {
    return editor;
  }

  /**
   * @return the menuBar
   */
  public JMenuBar getMenuBar() {
    return menuBar;
  }

  private JMenu createFileMenu() {
    JMenu menu = new JMenu(Reader.
            getPropertyValueAsString(Reader.StringProperty.labelFileMenu));

    /**
     * New Menu Item *
     */
    newImage = new JMenuItem(Reader.getPropertyValueAsString(
            Reader.StringProperty.labelNew));
    newImage.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
                ActionPerformedEvent.Action.newImage));
      }
    });
    newImage.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_N,
                                               ActionEvent.CTRL_MASK));
    menu.add(newImage);


    /**
     * Open Menu Item *
     */
    openImage = new JMenuItem(Reader.getPropertyValueAsString(
            Reader.StringProperty.labelOpen));
    openImage.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
                ActionPerformedEvent.Action.openImage));
      }
    });
    openImage.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_O,
                                               ActionEvent.CTRL_MASK));
    menu.add(openImage);
    
    /*
    menu.add(new JMenu(Reader.getPropertyValueAsString(
            Reader.StringProperty.labelLastOpened)));*/
    
    /**
     * Save Menu Item *
     */
    saveImage = new JMenuItem(Reader.getPropertyValueAsString(
            Reader.StringProperty.labelSave));
    saveImage.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
                ActionPerformedEvent.Action.saveImage));
      }
    });
    saveImage.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_S,
                                               ActionEvent.CTRL_MASK));
    saveImage.setEnabled(false);
    menu.add(saveImage);
    
    
    /**
     * SaveAs Menu Item *
     */
    saveImageAs = new JMenuItem(Reader.getPropertyValueAsString(
            Reader.StringProperty.labelSaveAs));
    saveImageAs.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
                ActionPerformedEvent.Action.saveImageAs));
      }
    });
    saveImageAs.setEnabled(false);
    menu.add(saveImageAs);
    menu.addSeparator();
    
    /**
     * InsertImage Menu Item *
     */
    insertImage = new JMenuItem(Reader.getPropertyValueAsString(
            Reader.StringProperty.labelInsertImage));
    insertImage.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
                ActionPerformedEvent.Action.insertImage));
      }
    });

    menu.add(insertImage);
    menu.addSeparator();
    
    /**
     * Print Menu Item *
     */
    printImage = new JMenuItem(Reader.getPropertyValueAsString(
            Reader.StringProperty.labelPrint));
    printImage.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
                ActionPerformedEvent.Action.print));
      }
    });
    printImage.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_P, ActionEvent.CTRL_MASK));
    printImage.setEnabled(false);
    menu.add(printImage);
    
    menu.addSeparator();

    
    /**
     * Quit Menu Item *
     */
    exitApplication = new JMenuItem(Reader.getPropertyValueAsString(
            Reader.StringProperty.labelQuit));
    exitApplication.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
            ActionPerformedEvent.Action.closeWindow));
      }
    });
    menu.add(exitApplication);

    return menu;
  }

  private JMenu createEditMenu() {
    JMenu menu = new JMenu(Reader.
            getPropertyValueAsString(Reader.StringProperty.labelEditMenu));
    JMenuItem item = null;

    /**
     * Undo Menu Item *
     */
    undo = new JMenuItem(Reader.getPropertyValueAsString(
            Reader.StringProperty.labelUndo));
    undo.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
                ActionPerformedEvent.Action.undo));
      }
    });
    undo.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Z, ActionEvent.CTRL_MASK));
    undo.setEnabled(false);
    menu.add(undo);
    
    /**
     * Redo Menu Item *
     */
    redo = new JMenuItem(Reader.getPropertyValueAsString(
            Reader.StringProperty.labelRedo));
    redo.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
                ActionPerformedEvent.Action.redo));
      }
    });
    redo.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Y, ActionEvent.CTRL_MASK));
    redo.setEnabled(false);
    menu.add(redo);
    menu.addSeparator();

    
    /**
     * Delete Menu Item *
     */
    delete = new JMenuItem(Reader.getPropertyValueAsString(
            Reader.StringProperty.labelDelete));
    delete.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
                ActionPerformedEvent.Action.delete));
      }
    });
    delete.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
    delete.setEnabled(false);
    menu.add(delete);

    /**
     * Cut Menu Item *
     */
    cut = new JMenuItem(Reader.getPropertyValueAsString(
            Reader.StringProperty.labelCut));
    cut.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
                ActionPerformedEvent.Action.cut));
      }
    });
    cut.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_X, ActionEvent.CTRL_MASK));
    cut.setEnabled(false);
    menu.add(cut);
    
    /**
     * Copy Menu Item *
     */
    copy = new JMenuItem(Reader.getPropertyValueAsString(
            Reader.StringProperty.labelCopy));
    copy.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
                ActionPerformedEvent.Action.copy));
      }
    });
    copy.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C, ActionEvent.CTRL_MASK));
    copy.setEnabled(false);
    menu.add(copy);
    
    /**
     * Paste Menu Item *
     */
    paste = new JMenuItem(Reader.getPropertyValueAsString(
            Reader.StringProperty.labelPaste));
    paste.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
                ActionPerformedEvent.Action.paste));
      }
    });
    paste.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_V, ActionEvent.CTRL_MASK));
    menu.add(paste);
   
    menu.addSeparator();
    
    
    /**
     * Duplicate Menu Item *
     */
    duplicate = new JMenuItem(Reader.getPropertyValueAsString(
            Reader.StringProperty.labelDuplicate));
    duplicate.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
                ActionPerformedEvent.Action.duplicate));
      }
    });
    //item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_A, ActionEvent.CTRL_MASK));
    duplicate.setEnabled(false);
    menu.add(duplicate);
    
    /**
     * TakePicture Menu Item *
     */
    takePicture = new JMenuItem(Reader.getPropertyValueAsString(
            Reader.StringProperty.labelTakePicture));
    takePicture.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
                ActionPerformedEvent.Action.takePicture));
      }
    });
    //item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_A, ActionEvent.CTRL_MASK));
    takePicture.setEnabled(false);
    //menu.add(takePicture);
    
    
    /**
     * Alignment Menu *
     */
    alignment = new JMenu(Reader.getPropertyValueAsString(Reader.StringProperty.labelAlignment));
    AlignmentMenuItem ami = new AlignmentMenuItem();
    alignment.add(ami);
    alignment.getInsets().set(0, 0, 0, 0);
    alignment.setComponentOrientation(ComponentOrientation.LEFT_TO_RIGHT);
    alignment.getMargin().set(0, 0, 0, 0);
    alignment.setVerticalAlignment(SwingConstants.TOP);
    alignment.setHorizontalAlignment(SwingConstants.LEFT);
    alignment.setEnabled(false);
    menu.add(alignment);
    
    
    /**
     * Bring To Front
     */
    bringToFront = new JMenuItem(Reader.getPropertyValueAsString(
            Reader.StringProperty.labelBringToFront));
    bringToFront.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
                ActionPerformedEvent.Action.bringToFront));
      }
    });
    bringToFront.setEnabled(false);
    menu.add(bringToFront);
    
    /**
     * Send To Back
     */
    sendToBack = new JMenuItem(Reader.getPropertyValueAsString(
            Reader.StringProperty.labelSendToBack));
    sendToBack.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
                ActionPerformedEvent.Action.sendToBack));
      }
    });
    sendToBack.setEnabled(false);
    menu.add(sendToBack);
    
    menu.addSeparator();
    
    
    /**
     * Group Menu Item *
     */
    group = new JMenuItem(Reader.getPropertyValueAsString(
            Reader.StringProperty.labelGroup));
    group.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
                ActionPerformedEvent.Action.group));
      }
    });
    group.setEnabled(false);
    menu.add(group);

    /**
     * Ungroup Menu Item *
     */
    ungroup = new JMenuItem(Reader.getPropertyValueAsString(
            Reader.StringProperty.labelUngroup));
    ungroup.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
                ActionPerformedEvent.Action.ungroup));
      }
    });
    //item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_I, ActionEvent.CTRL_MASK));
    ungroup.setEnabled(false);
    menu.add(ungroup);

    menu.addSeparator();

    /**
     * MagneticGrid Menu Item *
     */
    magGrid = new JMenu(Reader.getPropertyValueAsString(Reader.StringProperty.labelMagGrid));
    showMagGrid = new JMenuItem(Reader.getPropertyValueAsString(Reader.StringProperty.labelShowMagGrid));
    showMagGrid.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
                ActionPerformedEvent.Action.showMagneticGrid,true));
      }
    });
    magGrid.add(showMagGrid);
    hideMagGrid = new JMenuItem(Reader.getPropertyValueAsString(Reader.StringProperty.labelHideMagGrid));
    hideMagGrid.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
                ActionPerformedEvent.Action.showMagneticGrid,false));
      }
    });
    magGrid.add(hideMagGrid);
    menu.add(magGrid);
    
    menu.addSeparator();
    
    /**
     * SelectAll Menu Item *
     */
    selectAll = new JMenuItem(Reader.getPropertyValueAsString(
            Reader.StringProperty.labelSelectAll));
    selectAll.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
                ActionPerformedEvent.Action.selectAll));
      }
    });
    selectAll.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_A,
                                               ActionEvent.CTRL_MASK));
    selectAll.setEnabled(false);
    menu.add(selectAll);

    /**
     * InvertSelection Menu Item *
     */
    invertSelection = new JMenuItem(Reader.getPropertyValueAsString(
            Reader.StringProperty.labelInvertSelection));
    invertSelection.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
                ActionPerformedEvent.Action.invertSelection));
      }
    });
    invertSelection.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_I,
                                               ActionEvent.CTRL_MASK));
    invertSelection.setEnabled(false);
    menu.add(invertSelection);

    /**
     * Deselect Menu Item *
     */
    deselect = new JMenuItem(Reader.getPropertyValueAsString(
            Reader.StringProperty.labelDeselect));
    deselect.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
                ActionPerformedEvent.Action.deselectAll));
      }
    });
    deselect.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_D,
                                               ActionEvent.CTRL_MASK));
    deselect.setEnabled(false);
    menu.add(deselect);
    
    return menu;
  }

  @Override
  public void elementManagerStateChanged(ElementManagerStateEvent e) {
    switch(e.getType()) {
      case hasSelectedElements:
        this.delete.setEnabled(e.isState());
        this.cut.setEnabled(e.isState());
        this.copy.setEnabled(e.isState());
        this.duplicate.setEnabled(e.isState());
        this.takePicture.setEnabled(e.isState());
        this.group.setEnabled(e.isState());
        this.ungroup.setEnabled(e.isState());
        this.invertSelection.setEnabled(e.isState());
        this.deselect.setEnabled(e.isState());
        this.bringToFront.setEnabled(e.isState());
        this.sendToBack.setEnabled(e.isState());
        break;
      case hasElements:
        this.saveImage.setEnabled(e.isState());
        this.saveImageAs.setEnabled(e.isState());
        this.printImage.setEnabled(e.isState());
        
        this.duplicate.setEnabled(e.isState());
        this.takePicture.setEnabled(e.isState());
        
        this.selectAll.setEnabled(e.isState());
        
        this.alignment.setEnabled(e.isState());
        if(!e.isState()) {
          this.delete.setEnabled(e.isState());
          this.cut.setEnabled(e.isState());
          this.copy.setEnabled(e.isState());
          this.invertSelection.setEnabled(e.isState());
          this.deselect.setEnabled(e.isState());
          this.group.setEnabled(e.isState());
          this.ungroup.setEnabled(e.isState());
          this.bringToFront.setEnabled(e.isState());
          this.sendToBack.setEnabled(e.isState());
        }
        break;
      case canRedo:
        this.redo.setEnabled(e.isState());
        break;
      case canUndo:
        this.undo.setEnabled(e.isState());
        break;
      case isInCompositeEditMode:
        this.compositeLeaveButton(e.isState());
        break;
    }
  }

  @Override
  public void canvasStateChanged(CanvasStateEvent e) {
    if(!e.getSelectedElements().isEmpty()) { // Selected Elems
      this.mousePosition.setText(e.getBoundingBox().x + "|" + e.getBoundingBox().y);
      this.delta.setText(e.getBoundingBox().width + "|" + e.getBoundingBox().height);
    }
    if(e.getMouseEvent()!=null) { // Mouse moved
      this.mousePosition.setText(e.getMouseEvent().getPoint().x + "|" + e.getMouseEvent().getPoint().y);
      if(e.getSelectedElements().isEmpty())
        this.delta.setText(e.getDelta().x + "|" + e.getDelta().y);
    } else {
      if(e.getSelectedElements().isEmpty()) { // Mouse dragged
        this.mousePosition.setText(e.getStartPoint().x + "|" + e.getStartPoint().y);
        this.delta.setText(e.getDelta().x + "|" + e.getDelta().y);
      } else if(e.getEndPoint().distance(new Point2D.Float(0,0)) > 0)
        this.mousePosition.setText(e.getBoundingBox().x + "|" + e.getBoundingBox().y);
        this.delta.setText(e.getBoundingBox().width + "|" + e.getBoundingBox().height);
    }
  }

  private void compositeLeaveButton(boolean state) {
    if(state) {
      if(!this.menuBar.getComponents()[this.menuBar.getComponents().length-1].equals(this.leaveGroup)) {
        leaveGroup = new JButton(Reader.getPropertyValueAsString(Reader.StringProperty.labelLeaveGroup));
        leaveGroup.addActionListener(new ActionListener() {
          @Override
          public void actionPerformed(ActionEvent e) {
            ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(
                    ActionPerformedEvent.Action.leaveGroup));
          }
        });
        this.leaveGroup.setPreferredSize(new Dimension(this.leaveGroup.getWidth(),16));
        this.menuBar.add(leaveGroup);
      }
    } else {
      if(leaveGroup != null)
        this.menuBar.remove(leaveGroup);
    }
    this.menuBar.getParent().repaint();
  }
}
