 /**----------------------------------------------------
 * File:      SVGComposite.java
 * Package:   at.hoertler.svgeditor.svg.elements
 * Project:   SVGEditor
 * ----------------------------------------------------
 * @author Philipp Hörtler BSc
 * @version 1.0.0
 * Date: 2013-03-24
 * ----------------------------------------------------
 * Description:
 * 
 * Represents a group of SVGElements.
 */
package at.hoertler.svgeditor.svg.elements;

import at.hoertler.svgeditor.common.HelperFunction;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.geom.Line2D;
import java.util.Vector;
import javax.swing.Icon;

public class SVGComposite
        extends SVGElement {

  private Vector<SVGElement> elements = new Vector<SVGElement>();
  private Vector<SVGElement> origElements = new Vector<SVGElement>();
  private Rectangle origBoundingRect = new Rectangle();
  private boolean inCompositeEditMode = false;

  public SVGComposite() {
    this.getHandles().add(new Point(-1,-1));
    this.getHandles().add(new Point(-1,-1));
    this.getHandles().add(new Point(-1,-1));
    this.getHandles().add(new Point(-1,-1));
    
  }


  @Override
  public SVGElement getInstance() {
    return new SVGComposite();
  }

  /**
   * @return the inCompositeEditMode
   */
  public boolean isInCompositeEditMode() {
    return inCompositeEditMode;
  }

  /**
   * @param inCompositeEditMode the inCompositeEditMode to set
   */
  public void setInCompositeEditMode(boolean inCompositeEditMode) {
    if(!inCompositeEditMode && inCompositeEditMode!=this.inCompositeEditMode){
      Point pMin = new Point(Integer.MAX_VALUE,Integer.MAX_VALUE);
      Point pMax = new Point(Integer.MIN_VALUE,Integer.MIN_VALUE);
      Dimension d = new Dimension(0,0);
      for(SVGElement elem : this.origElements) {
        elem.deSelect();
        for(Point h : elem.getHandles()) {
          if(pMin.x>h.x) pMin.x = h.x;
          if(pMin.y>h.y) pMin.y = h.y;
          if(pMax.x<h.x) pMax.x = h.x;
          if(pMax.y<h.y) pMax.y = h.y;
        }
      }
      d.height = Math.abs(pMax.y-pMin.y);
      d.width = Math.abs(pMax.x-pMin.x);
      Rectangle newOBB = new Rectangle(pMin,d);
      Rectangle bb = (Rectangle)this.getBoundingBox().clone();
      double scaleW = bb.width/(double)this.origBoundingRect.width;
      double scaleH = bb.height/(double)this.origBoundingRect.height;
      Point trans = HelperFunction.calculateTranslationPoint(this.origBoundingRect.getLocation(), newOBB.getLocation());
      trans.x *= scaleW;
      trans.y *= scaleH;
      this.getBoundingBox().height = (int)(this.getBoundingBox().getHeight()/this.origBoundingRect.getHeight()*newOBB.getHeight());
      this.getBoundingBox().width = (int)(this.getBoundingBox().getWidth()/this.origBoundingRect.getWidth()*newOBB.getWidth());
      this.getBoundingBox().x += trans.x;
      this.getBoundingBox().y += trans.y;
      this.origBoundingRect = newOBB;
      recalculateElements();
      this.recalculateHandles();
    }
    this.inCompositeEditMode = inCompositeEditMode;
    
  }

  @Override
  public void paint(Graphics g) {
    this.paint(g, 1);
  }

  @Override
  public void paint(Graphics g, int zoomFactor) {
    for (SVGElement e : this.getElements())
      e.paint(g, zoomFactor);
  }

  @Override
  public Icon getIcon() {
    return new Icon() {
      public int getIconHeight() {
        return 16;
      }

      public int getIconWidth() {
        return 16;
      }

      @Override
      public void paintIcon(Component c, Graphics g, int x, int y) {
        Graphics2D g2 = (Graphics2D) g;
        g2.setColor(Color.BLACK);
        g2.setStroke(new BasicStroke(2, BasicStroke.CAP_BUTT,
                                     BasicStroke.JOIN_MITER));
        g2.drawLine(x + 2, y + 2, x + 14, y + 14);
        g2.drawLine(x + 5, y + 2, x + 14, y + 9);
      }
    };
  }

  @Override
  public void moveHandle(int handleIndex, Point p, MouseEvent e) {
    if (this.isNormalMode())
      return;
    for(SVGElement elem : this.origElements) {
      if(elem.getClass().isAssignableFrom(SVGText.class))
        return;
    }
    this.getHandles().get(handleIndex).translate(p.x, p.y);
    int leftH = (handleIndex + 3) % 4;
    int rightH = (handleIndex + 5) % 4;
    Point h1 = this.getHandles().get(leftH);
    h1.x = this.getHandles().get((leftH + 5) % 4).x;
    h1.y = this.getHandles().get((leftH + 3) % 4).y;
    Point h2 = this.getHandles().get(rightH);
    h2.x = this.getHandles().get((rightH + 5) % 4).x;
    h2.y = this.getHandles().get((rightH + 3) % 4).y;

    //this.recalculateHandles();
    this.updateBoundingRect(this.getHandles());
    recalculateElements();
  }

  @Override
  public boolean validate() {
    for (SVGElement e : this.getElements())
      if (!e.validate())
        return false;

    return true;
  }
  
  @Override
  protected void calculateNormalFromEdit() {
    if(this.validate()) {
      this.normalHandles.clear();
      this.normalHandles.addAll((Vector<Point>)this.editHandles.clone());
      this.editHandles.clear();
    } else {
      this.normalHandles.clear();
      Point p = this.origBoundingRect.getLocation();
      Dimension d = this.origBoundingRect.getSize();
      this.normalHandles.add(p);
      this.normalHandles.add(new Point(p.x+d.width, p.y));
      this.normalHandles.add(new Point(p.x+d.width, p.y+d.height));
      this.normalHandles.add(new Point(p.x, p.y+d.height));
    }
    this.updateBoundingRect(this.normalHandles);
    this.mode = Mode.Normal;
    this.recalculateElements();
  }

  @Override
  public String exportToSVG() {
    StringBuilder sb = new StringBuilder();
    sb.append("<g id=\"" + this.getClass().getSimpleName() + "\">\n");
    for (SVGElement e : this.getElements())
      sb.append(e.exportToSVG());
    sb.append("</g>\n");
    return sb.toString();
  }

  @Override
  public void handleKeyPressed(KeyEvent e) {
    throw new UnsupportedOperationException("Not supported yet.");
  }

  @Override
  public void handleKeyReleased(KeyEvent e) {
    throw new UnsupportedOperationException("Not supported yet.");
  }

  @Override
  public void handleKeyTyped(KeyEvent e) {
    throw new UnsupportedOperationException("Not supported yet.");
  }

  @Override
  public void handleMouseMoved(MouseEvent e) {
    //throw new UnsupportedOperationException("Not supported yet.");
  }

  @Override
  public void handleMouseEntered(MouseEvent e) {
    //throw new UnsupportedOperationException("Not supported yet.");
  }

  @Override
  public void handleMouseExited(MouseEvent e) {
    //throw new UnsupportedOperationException("Not supported yet.");
  }

  @Override
  public void handleMousePressed(MouseEvent e) {
    // this.getHandles().set(0, e.getPoint());
    //this.getHandles().set(2, e.getPoint());
    this.updateBoundingRect(this.getHandles());
  }

  @Override
  public void handleMouseDragged(MouseEvent e) {
    //this.getHandles().set(2, e.getPoint());
    //this.updateBoundingRect(this.getHandles());
  }

  @Override
  public void handleMouseReleased(MouseEvent e) {
    //this.getHandles().set(1, e.getPoint());
    //this.updateBoundingRect(this.getHandles());
  }

  @Override
  public void handleMouseClicked(MouseEvent e) {
    
  }

  @Override
  public boolean hit(Rectangle rect) {
    Vector<Point> handles = this.getHandles();
    Line2D l = new Line2D.Double(handles.get(0), handles.get(1));
    return this.getBoundingBox().intersects(rect);
  }

  /**
   * @return the elements
   */
  public Vector<SVGElement> getElements() {
    if(this.isInCompositeEditMode())
      return this.origElements;
    return elements;
  }

  /**
   * @param elements the elements to set
   */
  public void setElements(Vector<SVGElement> elements) {
    this.elements.clear();
    for(SVGElement e : elements) 
      this.addElement(e);
  }

  public void addElement(SVGElement elem) {
    //ZZ this.getElements().add(elem);
    this.elements.add(elem);

    this.recalculateHandles();

    this.updateBoundingRect(this.getHandles());
    this.origElements.clear();
    for (SVGElement e : this.elements) {//ZZ this.getElements()) 
      SVGElement el = e.clone();
      el.setMode(e.getMode());
      e.deSelect();
      this.origElements.add(e.clone());
    }
    this.origBoundingRect = (Rectangle) this.getBoundingBox().clone();
  }

  @Override
  public void translate(int x, int y) {
    super.translate(x, y);
    if (this.isNormalMode())
      this.setMode(Mode.Edit);
    
    /*
     * for(SVGElement elem : this.getElements()) elem.translate(x,y);
     *
     * this.updateBoundingRect(this.getHandles());
     */
    recalculateElements();
  }

  private void recalculateHandles() {
    Vector<Point> handlesTmp = (Vector<Point>) this.getHandles().clone();
    this.getHandles().clear();
    for (SVGElement e : this.elements)
      this.getHandles().addAll(e.getHandles());

    Point h0 = new Point(minX(this.getHandles(), 1), minY(this.getHandles(), 1));
    Point h1 = new Point(maxX(this.getHandles(), 1), minY(this.getHandles(), 1));
    Point h2 = new Point(maxX(this.getHandles(), 1), maxY(this.getHandles(), 1));
    Point h3 = new Point(minX(this.getHandles(), 1), maxY(this.getHandles(), 1));
    this.getHandles().clear();
    this.getHandles().addAll(handlesTmp);
    this.getHandles().get(0).setLocation(h0);
    this.getHandles().get(1).setLocation(h1);
    this.getHandles().get(2).setLocation(h2);
    this.getHandles().get(3).setLocation(h3);
  }

  @Override
  public SVGElement clone() {
    
    SVGComposite comp = (SVGComposite) super.clone();
    for (SVGElement elem : this.origElements) {
      SVGElement e = elem.clone();
      comp.origElements.add(e);
    }
    comp.origBoundingRect = (Rectangle)this.origBoundingRect.clone();
    //comp.setBoundingBox((Rectangle)this.getBoundingBox().clone());
    comp.elements.clear();
    /*for (SVGElement elem : this.getElements())
      comp.elements.add(elem.clone());*/
    comp.recalculateElements();
    return comp;
  }

  private void recalculateElements() {
    Rectangle oBB = this.origBoundingRect;
    Rectangle aBB = this.getBoundingBox();
    Point transBounding = HelperFunction.
            calculateTranslationPoint(oBB.getLocation(),
                                      aBB.getLocation());
    double scaleW = aBB.width/(double)oBB.width;
    double scaleH = aBB.height/(double)oBB.height;
    
    this.elements.clear();
    for(SVGElement elem : this.origElements) {
      SVGElement e = elem.clone();
      e.setMode(Mode.Edit);
      
      
      // Verzerrung
      
      if(scaleH != 1 || scaleW != 1) {
        e.getHandles().clear();
        for(int i = 0; i<elem.getHandles().size(); i++) {
          Point p = elem.getHandles().get(i);
          int x = (int)((p.x - oBB.x)*scaleW);
          int y = (int)((p.y - oBB.y)*scaleH);
          Point p1 = new Point(x,y);
          p1.translate(oBB.x, oBB.y);
          Point trans = HelperFunction.calculateTranslationPoint(p, p1);
          e.getHandles().add(new Point(p1.x,p1.y));
          //e.moveHandle(i, p1, null);
        }
      }

      // Verschiebung
      e.translate(transBounding);
      e.setMode(Mode.Normal);
      this.elements.add(e);
    }
  }

  @Override
  public Point getTranslationPointForHandle(int handleIndex) {
    return this.getHandles().get((handleIndex+2)%4);
  }
}
