 /**----------------------------------------------------
 * File:      SVGElement.java
 * Package:   at.hoertler.svgeditor.svg.elements
 * Project:   SVGEditor
 * ----------------------------------------------------
 * @author Philipp Hörtler BSc
 * @version 1.0.0
 * Date: 2013-03-24
 * ----------------------------------------------------
 * Description:
 * 
 * Is the Baseclass of all SVGElements and can be used to create new ones.
 */
package at.hoertler.svgeditor.svg.elements;

import at.hoertler.svgeditor.action.ActionManager;
import at.hoertler.svgeditor.common.KeyInputManager;
import at.hoertler.svgeditor.event.ActionPerformedEvent;
import at.hoertler.svgeditor.event.OptionChangedEvent;
import at.hoertler.svgeditor.options.OptionManager;
import at.hoertler.svgeditor.options.OptionManager.Alignment;
import at.hoertler.svgeditor.options.OptionManager.TextOption;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.Stroke;
import java.awt.Toolkit;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.font.GlyphVector;
import java.awt.font.TextAttribute;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.text.AttributedString;
import java.util.ArrayList;
import java.util.Vector;
import javax.swing.Icon;

public abstract class SVGElement {


  public enum Mode { InitialCreate, ExtendedCreate, Edit, ExtendedEdit, Normal , TextEdit};
  
  private int handleClickTolerance = 2;
  protected int standardTextMargin = 5;
  
  // Properties
  private int lineStrength;
  private boolean fillcolorSetable = false;
  private Color fillColor;
  private Color lineColor;
  private boolean textSetable = false;
  private String text = "";
  private Font font;
  private Color textColor;
  private int textSize;
  private Alignment textAlignment;
  private ArrayList<TextOption> textOption = new ArrayList<TextOption>();
  
  private Graphics2D graphics;
  protected Mode mode = Mode.InitialCreate;
  protected Vector<Point> initHandles = new Vector<Point>();
  protected Vector<Point> editHandles = new Vector<Point>();
  protected Vector<Point> extendedHandles = new Vector<Point>();
  protected Vector<Point> normalHandles = new Vector<Point>();
  
  protected Rectangle initBoundingBox = null;
  protected Rectangle normalBoundingBox = null;
  protected Rectangle textBoundingBox = null;
  


  /**
   * This method creates an AttributedString with the Textoptions set for this SVGElement.
   * @param txt String-Text to be attributed.
   * @param zoomFactor zoomFactor to adjust the set TextSize.
   * @return The AttributedString Object.
   */
  public AttributedString getAttributedString(String txt, int zoomFactor) {
    AttributedString as = new AttributedString("");
    this.getGraphics().setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                                        RenderingHints.VALUE_ANTIALIAS_ON);
    Font plainFont = new Font(this.getFont().getName(), Font.TRUETYPE_FONT, this.
            getTextSize()*zoomFactor);
    if(txt.isEmpty())
      txt+=" ";
    as = new AttributedString(txt);
    as.addAttribute(TextAttribute.FOREGROUND, this.getTextColor());
    
    int textOption = 0;
    for (OptionManager.TextOption o : this.getTextOption()) {
      switch (o) {
        case italic:
          textOption |= Font.ITALIC;
          break;
        case bold:
          textOption |= Font.BOLD;
          break;
        case underlined:
          as.
                  addAttribute(TextAttribute.UNDERLINE,
                               TextAttribute.UNDERLINE_ON, 0, txt.length());
          break;
        case strikethrough:
          as.addAttribute(TextAttribute.STRIKETHROUGH,
                          TextAttribute.STRIKETHROUGH_ON, 0, txt.length());
          break;
      }
    }
    as.addAttribute(TextAttribute.FONT, plainFont.deriveFont(textOption));
    return as;
  }
  
  public void renderText(int zoomFactor) {
    //this.setTextMargin((Math.abs(this.getLineStrength()) + this.standardTextMargin)*zoomFactor);
    
    if (this.getText() != "" || this.isTextEditMode()) {
      String text = this.getText();
      if(text.isEmpty()) text+=" ";
      if(text.endsWith("\n"))
        text+=" ";
      text+="Jj";
      
      String[] lines = text.split("\n");
      int height = 0;
      int width = 0;
      AttributedString as = this.getAttributedString(text, 1);
      Font f = (Font)as.getIterator().getAttribute(TextAttribute.FONT);
      for(int i = 0; i<lines.length; i++) {
      GlyphVector vec = f.createGlyphVector(this.getGraphics().getFontRenderContext(), lines[i]);
      //height = Math.max(height, (int)Math.ceil(vec.getVisualBounds().getHeight()));
      width = Math.max(width, (int)Math.ceil(vec.getVisualBounds().getBounds().width));
      }
      height = this.getTextSize();
      lines[lines.length-1] = lines[lines.length-1].substring(0, lines[lines.length-1].length()-2);
      
      
      for(int i = 0; i < lines.length; i++) {
        if(lines[i].length()==0)
          lines[i] += " ";
        as = this.getAttributedString(lines[i], zoomFactor);
        f = (Font)as.getIterator().getAttribute(TextAttribute.FONT);
        Rectangle stringBounds = this.getGraphics().getFontMetrics(f).getStringBounds(lines[i], this.getGraphics()).getBounds();
        int x = 0;
        switch(this.getTextAlignment()) {
          case left:
            x = this.getTextBoundingBox().x;
            break;
          case center:
            x = this.getTextBoundingBox().x + this.getTextBoundingBox().width / 2 - ((stringBounds.width)/zoomFactor) / 2;
            break;
          case right:
            x = this.getTextBoundingBox().x + this.getTextBoundingBox().width - ((stringBounds.width)/zoomFactor);
            break;
        }
        int y = this.getTextBoundingBox().y + this.getTextBoundingBox().height / 2 - (lines.length * height) / 2 + (i+1)*height;
        if(i==0) y-=4;
        Point txtPosition = calcZoom(new Point(x, y), zoomFactor);
        this.getGraphics().drawString(as.getIterator(), txtPosition.x, txtPosition.y);
        if(i==lines.length-1 && this.isTextEditMode()) {
          GlyphVector vec = f.createGlyphVector(this.getGraphics().getFontRenderContext(), lines[i]);
          int w = Math.max(width, (int)Math.ceil(vec.getVisualBounds().getBounds().width));
          this.getGraphics().setColor(Color.BLUE);
          this.getGraphics().fillRect(txtPosition.x + stringBounds.width, txtPosition.y-height, 3, height);
          //this.getGraphics().drawString(as.getIterator(), txtPosition.x, txtPosition.y);
        }
      }
      if(lines.length==0) {
      Point txtPosition = this.getTextBoundingBox().getLocation();
      Dimension dim = this.getTextBoundingBox().getSize();
      GlyphVector vec = f.createGlyphVector(this.getGraphics().getFontRenderContext(), "Jj");
      int w = Math.max(width, (int)Math.ceil(vec.getVisualBounds().getBounds().width));
      this.getGraphics().setColor(Color.BLUE);
      this.getGraphics().fillRect(txtPosition.x, txtPosition.y, dim.width, dim.height);
      }
     }
   }

  /**
   * Method getInstance.
   *
   * @return SVGElement
   */
  public abstract SVGElement getInstance();

  /**
   * Paints the element to the given Graphics-Object
   *
   * @param g Graphics object
   */
  public abstract void paint(Graphics g);

  /**
   * Paints the element to the given Graphics-Object
   *
   * @param g          Graphics object
   * @param zoomFactor zoomFactor
   */
  public abstract void paint(Graphics g, int zoomFactor);

  /**
   * @return the stroke object for the given lineStrength.
   */
  protected Stroke getStroke(int zoomFactor) {
    Stroke stroke = null;
    int strength = 0;
    double ls = this.getExportLineStrength();
    if(ls==0.25) {
      strength = 1 * zoomFactor / 4;
      if(strength==0)
        strength = 1;
    } else
      strength = (int)ls*zoomFactor;
    if(strength == 0 && !this.isFillcolorSetable())
      strength = 1;
    stroke = new BasicStroke(strength,
                               BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND);
    return stroke;
  }

  /**
   * Paints all handles of this element.
   *
   * @param g Graphics object used for painting
   */
  public void paintHandles(Graphics g) {
    this.paintHandles(g, 1);
  }

  /**
   * Paints all handles of this element.
   *
   * @param g          Graphics object used for painting
   * @param zoomFactor Zoomfactor to be used
   */
  public void paintHandles(Graphics g, int zoomFactor) {
    if (this.isNormalMode()) {
      return;
    }
    for(int i = 0; i<this.getHandles().size(); i++)
      this.paintHandle(g, i, zoomFactor);
  }
  
   /**
   * Paints the handle for the given handleIndex of this element.
   *
   * @param g          Graphics object used for painting
   * @param handleIndex the index of the handle to be painted
   * @param zoomFactor Zoomfactor to be used
   */
  public void paintHandle(Graphics g, int handleIndex, int zoomFactor) {
    ((Graphics2D) g).setStroke(new BasicStroke(1, BasicStroke.CAP_BUTT,
                                               BasicStroke.JOIN_MITER));
    Rectangle2D r = new Rectangle2D.Double();
    Point hdl = this.getHandles().get(handleIndex);
    Point p = new Point(hdl.x * zoomFactor, hdl.y * zoomFactor);
      r.setRect(p.x - handleClickTolerance, p.y - handleClickTolerance,
                2 * handleClickTolerance, 2 * handleClickTolerance);
      g.setColor(Color.WHITE);
      ((Graphics2D) g).fill(r);
      g.setColor(Color.BLACK);
      ((Graphics2D) g).draw(r);
  }

  /**
   * Checks if the given point hits the element.
   *
   * @param point Point to be checked for hitting the element.
   *
   * @return true if the element is hit.
   */
  public boolean hit(Point point) {
    point.translate(-1, -1);
    return this.hit(point, new Dimension(2, 2));
  }

  /**
   * Checks if the given rect - discribed by parameters for point (upper left
   * corner) and dimension) - hits the element.
   *
   * @param point Point for the upper left corner.
   * @param d     Dimension
   *
   * @return true if the element is hit.
   */
  public boolean hit(Point point, Dimension d) {
    return this.hit(new Rectangle(point, d));
  }

  /**
   * Checks if the given rectangle hits the element.
   *
   * @param rect Rectangel to be checked.
   *
   * @return true if the element is hit.
   */
  public abstract boolean hit(Rectangle rect);

  /**
   * Checks if the given point hits the element.
   *
   * @param point Point to be checked for hitting the element.
   *
   * @return true if the element is hit.
   */
  public boolean hitHandle(Point point) {
    if (this.isNormalMode()) {
      return false;
    }
    return (this.getHandleIndex(point) == -1) ? false : true;
  }

  /**
   * Gets the handle point.
   *
   * @param p Point that should be checked for a handle.
   *
   * @return The point of the handle that has been found or null.
   */
  public Point getHandleForPoint(Point p) {
    Vector<Point> handles = this.getHandles();
    int i = this.getHandleIndex(p);
    if (i != -1) {
      return handles.get(i);
    }
    return null;
  }

  /**
   * Gets the handle index if one is hit.
   *
   * @param p Point to get the handle index for.
   *
   * @return int Handle index. -1 if no handle has been hit.
   */
  public int getHandleIndex(Point p) {
    Vector<Point> handles = this.getHandles();
    for (int i = 0; i < handles.size(); i++) {
      Point point = handles.get(i);
      if (p.distance(point) <= this.handleClickTolerance) {
        return i;
      }
    }
    return -1;
  }

  /**
   * This method creates an icon for an specific element.
   *
   * @return The created icon.
   */
  public abstract Icon getIcon();
  
  /**
   * This method gets the related Point to the given handle index needed for a proportion keeping translation of rectangular shapes.
   * In case of a rectangular shaped bounding box, this method returns the oposite handle index.
   * For different shapes override is neccessary e.g. Line, Arrow, Polygone, ...
   * 
   * @return The Point to translate the element with its proportion kept.
   */
  public abstract Point getTranslationPointForHandle(int handleIndex);

  /**
   * Moves the element for the given point.
   * This only works for selected Elements.
   *
   * @param p Point for whicht the element should be moved.
   */
  public void translate(Point p) {
    this.translate(p.x, p.y);
  }

  /**
   * Moves the element for the given x-/y-values.
   * This only works for selected Elements.
   *
   * @param x Value to move the element horizontally.
   * @param y Value to move the element vertically.
   */
  public void translate(int x, int y) {
    for (Point p : this.getHandles()) {
      p.translate(x, y);
    }
    this.updateBoundingRect(this.getHandles());
  }

  /**
   * Calculates the point with a given zoom factor.
   *
   * @param p          Original point
   * @param zoomFactor given zoom factor
   *
   * @return calculated zoomed point
   */
  protected Point calcZoom(Point p, int zoomFactor) {
    if(zoomFactor==1)
      return (Point)p.clone();
    return new Point(p.x * zoomFactor, p.y * zoomFactor);
  }

  /**
   * Moves an handle for a given index.
   *
   * @param handleIndex Index of the handle that should be moved.
   * @param p           Point to which the handle should be moved.
   */
  public abstract void moveHandle(int handleIndex, Point p, MouseEvent e);

  /**
   * @return the handles
   */
  public Vector<Point> getHandles() {
    switch(this.getMode()) {
      case Normal:
        return this.normalHandles;
      case InitialCreate:
        return initHandles;
      case ExtendedCreate:
        return initHandles;
      case Edit:
        return editHandles;
      case ExtendedEdit:
        return extendedHandles;
      case TextEdit:
        return this.editHandles;
      default:
        return this.normalHandles;
    }
  }

  /**
   * @param handles the handles to set
   */
  public void setHandles(Vector<Point> handles) {
    switch(this.getMode()) {
      case Normal:
        this.normalHandles = handles;
        break;
      case InitialCreate:
        this.initHandles = handles;
        break;
      case ExtendedCreate:
        this.initHandles = handles;
        break;
      case Edit:
        this.editHandles = handles;
        break;
      case ExtendedEdit:
        this.extendedHandles = handles;
        break;
      case TextEdit:
        this.normalHandles = handles;
        break;
      default:
        this.normalHandles = handles;
        break;
    }
  }

  


  /**
   * This method validates the actual element if it has been created correctly.
   * This method is called whenever the EditMode is left. If the validation
   * fails, the changes won't be saved.
   *
   * @return If the element is valid and can be rendered.
   */
  public abstract boolean validate();

  /**
   * Exports the element to an svg-string
   *
   * @return SVG string representation of the element
   */
  public abstract String exportToSVG();

  
  public String exportTextToSVG() {
    String[] lines = this.getText().split("\n");
    int x = (int)this.getTextBoundingBox().x;
    int y = (int)this.getTextBoundingBox().y + (this.getTextBoundingBox().height-(this.getTextSize()*lines.length))/2;
    Color c = this.getTextColor();
    StringBuilder sb = new StringBuilder();
    sb.append("\t\t");
    sb.append("<text ");
    sb.append("x=\"" + x + "\" "); // only needed for import! x-y is the upper-left corner of the text-bounding-box
    sb.append("y=\"" + y + "\" ");
    sb.append("fill=\"rgb("
            + c.getRed() + ","
            + c.getGreen() + ","
            + c.getBlue() + ")\" ");
    sb.append("font-family=\"");
    sb.append(this.getFont().getName());
    sb.append("\" ");
    sb.append("font-size=\"");
    sb.append(this.getTextSize());
    sb.append("\" ");
    sb.append("text-anchor=\"");
    switch(this.getTextAlignment()) {
      case left:
        sb.append("start");
        x = (int)this.getTextBoundingBox().x;
        break;
      case center:
        sb.append("middle");
        x = (int)this.getTextBoundingBox().x + (this.getTextBoundingBox().width/2);
        break;
      case right:
        sb.append("end");
        x = (int)this.getTextBoundingBox().x+this.getTextBoundingBox().width;
        break;
    }
    sb.append("\" ");
    if(this.getTextOption().contains(TextOption.bold))
      sb.append("font-weight=\"bold\" ");
    if(this.getTextOption().contains(TextOption.italic))
      sb.append("font-style=\"italic\" ");
    if(this.getTextOption().contains(TextOption.underlined) ||
       this.getTextOption().contains(TextOption.strikethrough)) {
      sb.append("text-decoration=\"");
      if(this.getTextOption().contains(TextOption.underlined))
        sb.append("underline");
      if(this.getTextOption().contains(TextOption.strikethrough))
        sb.append("line-through");
      sb.append("\" ");
    }
    sb.append(">\n");
    
    
    int offset = 1;
    for(int i = 0; i<lines.length; i++) {
      if(lines[i].isEmpty()) {
        offset++;
        continue;
      }
      sb.append("\t\t\t");
      sb.append("<tspan ");
      sb.append("x=\"" + x + "\" ");
      sb.append("dy=\"" + this.getTextSize()*offset + "\" ");
      
      sb.append(">");
      sb.append(lines[i]);
      sb.append("</tspan>\n");
      offset=1;
    }
    sb.append("\t\t");
    sb.append("</text>\n");
    // text-decoration = "line-through,underline"
    // font-style="italic" font-weight="bold"
    return sb.toString();
  }
  
  
  /**
   * This method sets the given Value for the given Option.
   * @param o Option to set.
   * @param value Value to set.
   */
  public void setValueForOption(OptionChangedEvent.Option o, Object value) {
    switch (o) {
      case lineStrength:
        setLineStrength((int) value);
        break;
      case lineColor:
        setLineColor((Color) value);
        break;
      case fillColor:
        setFillColor((Color) value);
        break;
      case font:
        setFont((Font) value);
        break;
      case text:
        setText((String) value);
        break;
      case textColor:
        setTextColor((Color) value);
        break;
      case textSize:
        setTextSize((int) value);
        break;
      case textAlignment:
        setTextAlignment((Alignment) value);
        break;
      case textOption:
        setTextOption((ArrayList<TextOption>)value);
        break;
      case zoomFactor:
        break;
      default:
        break;
    }
  }

  /**
   * Invoked when a key has been pressed.
   *
   * @param e KeyEvent.
   */
  public void handleKeyPressed(KeyEvent e) {
    switch(e.getKeyCode()) {
      case KeyEvent.VK_BACK_SPACE:
        if(this.getText().length()>0)
          this.setText(this.getText().substring(0, this.getText().length()-1));
      
        break;
    }
  }

  /**
   * Invoked when a key has been released.
   *
   * @param e KeyEvent.
   */
  public void handleKeyReleased(KeyEvent e) {
    switch(e.getKeyCode()) {
      case KeyEvent.VK_BACK_SPACE:
        break;
      case KeyEvent.VK_DELETE:
        break;
      case KeyEvent.VK_ENTER:
        if(!e.isShiftDown()) {
          this.setMode(Mode.Edit);
          this.setText(this.getText().substring(0, this.getText().length()-1));
        }
        break;
    }
  }

  /**
   * Invoked when a key has been Typed.
   *
   * @param e KeyEvent.
   */
  public void handleKeyTyped(KeyEvent e) {
    char c = e.getKeyChar();
    if((this.getFont().canDisplay(c) && c >=32 && c < 127 && c != 60 && c != 62) || c==10 || c==13 || c==196 || c==214 || c==220) {
      this.setText(this.getText()+c);
    }
  }

  /**
   * Invoked when the mouse has been moved.
   *
   * @param e MouseEvent
   */
  public void handleMouseMoved(MouseEvent e) { }

  /**
   * Invoked when the mouse entered a component.
   *
   * @param e MouseEvent
   */
  public void handleMouseEntered(MouseEvent e) { }

  /**
   * Invoked when the mouse left a component.
   *
   * @param e MouseEvent
   */
  public void handleMouseExited(MouseEvent e) { }

  /**
   * Invoked when the mouse has been pressed
   *
   * @param e MouseEvent
   */
  public abstract void handleMousePressed(MouseEvent e);

  /**
   * Invoked when the mouse has been dragged.
   *
   * @param e MouseEvent
   */
  public abstract void handleMouseDragged(MouseEvent e);

  /**
   * Invoked when the mouse has been Released
   *
   * @param e MouseEvent
   */
  public abstract void handleMouseReleased(MouseEvent e);

  /**
   * Invoked when the mouse has been clicked.
   *
   * @param e MouseEvent
   */
  public abstract void handleMouseClicked(MouseEvent e);

  /**
   * This method updates the bounding rect after it has been edited. Call this
   * method after every change you want to be painted immediately.
   *
   * @param handles Vector of handles are used to calculate the bounding rect.
   */
  public void updateBoundingRect(Vector<Point> handles) {
    int x1 = minX(handles, 1);
    int y1 = minY(handles, 1);
    int x2 = maxX(handles, 1);
    int y2 = maxY(handles, 1);

    Dimension d = new Dimension(x2 - x1, y2 - y1);


    // Create Shape
    if(this.isInitialCreateMode() || this.isExtendedCreateMode())
      this.initBoundingBox = new Rectangle(x1, y1, Math.abs(d.width), Math.abs(d.height));
    
    this.normalBoundingBox = new Rectangle(x1, y1, Math.abs(d.width), Math.abs(d.height));
    this.updateTextBoundingBox();
    
  }
  
  /**
   * This method is called after the bounding rectangle is updated.
   * This method sets the bounding rectangle for the text propertie of the SVGElement with the set margin.
   */
  public void updateTextBoundingBox() {
    this.setTextBoundingBox( new Rectangle(
                          this.getBoundingBox().x + this.getTextMargin(), 
                          this.getBoundingBox().y + this.getTextMargin(), 
                          this.getBoundingBox().width - this.getTextMargin()*2, 
                          this.getBoundingBox().height - this.getTextMargin()*2)
                        );
  }

  protected int minX(Vector<Point> handles, int zoomFactor) {
    int min = Integer.MAX_VALUE;
    for(Point p : handles) {
      p = calcZoom(p, zoomFactor);
      if(p.x<min)
        min = p.x;
    }
    return min;
  }
    
  protected int minY(Vector<Point> handles, int zoomFactor) {
    int min = Integer.MAX_VALUE;
    for (Point p : handles) {
      p = calcZoom(p, zoomFactor);
      if (p.y < min) {
        min = p.y;
      }
    }
    return min;
  }

  protected int maxX(Vector<Point> handles, int zoomFactor) {
    int min = Integer.MIN_VALUE;
    for (Point p : handles) {
      p = calcZoom(p, zoomFactor);
      if (p.x > min) {
        min = p.x;
      }
    }
    return min;
  }

  protected int maxY(Vector<Point> handles, int zoomFactor) {
    int min = Integer.MIN_VALUE;
    for (Point p : handles) {
      p = calcZoom(p, zoomFactor);
      if (p.y > min) {
        min = p.y;
      }
    }
    return min;
  }
  
  /**
   * This method is called when a handle has been moved and the operation has been finished.
   * e.g. Releasing the mousebutton after moving a handle around.
   * 
   * @param handleIndex The index of the handle that has been moved.
   */
  public void moveHandleFinished(int handleIndex, MouseEvent e) {}

  /**
   * Create a clone of the original SVGElement
   * @return the clone
   */
  public SVGElement clone() {
    SVGElement clone = this.getInstance();
    clone.setFillcolorSetable(this.isFillcolorSetable());
    clone.setFillColor(this.getFillColor());
    clone.setLineColor(this.getLineColor());
    clone.setLineStrength(this.getLineStrength());
    //clone.setBoundingBox(this.getBoundingBox());
    clone.mode = this.getMode();
    clone.setFont(this.getFont());
    clone.setTextSetable(this.isTextSetable());
    clone.setText(this.getText());
    clone.setTextAlignment(this.getTextAlignment());
    clone.setTextOption(this.getTextOption());
    clone.setTextSize(this.getTextSize());
    clone.setTextColor(this.getTextColor());
    clone.setHandles(new Vector<Point>());
    if(this.getGraphics() != null)
      clone.setGraphics((Graphics2D) this.getGraphics().create());
    
    //for (Point p : this.getHandles())
    //  clone.getHandles().add((Point)p.clone());
    
    clone.initHandles.clear();
    for (Point p : this.initHandles)
      clone.initHandles.add((Point)p.clone());
    
    clone.editHandles.clear();
    for (Point p : this.editHandles)
      clone.editHandles.add((Point)p.clone());
    
    clone.extendedHandles.clear();
    for (Point p : this.extendedHandles)
      clone.extendedHandles.add((Point)p.clone());
    
    clone.normalHandles.clear();
    for (Point p : this.normalHandles)
      clone.normalHandles.add((Point)p.clone());

    clone.initBoundingBox = (Rectangle)this.initBoundingBox.clone();
    clone.normalBoundingBox = (Rectangle)this.normalBoundingBox.clone();
    clone.textBoundingBox = (Rectangle)this.textBoundingBox.clone();
    if(!this.getClass().isAssignableFrom(SVGText.class)) {
      clone.setMode(Mode.Edit);
      clone.setMode(Mode.Normal);
    }
    return clone;
  }
  
  
  
  
  
  
 /*----------------------------------------------------------------------------
  * 
  * Get-/Set-Mode
  * Mode-Operations
  * 
  ----------------------------------------------------------------------------*/

  /**
   * @return selects the element
   */
  public void select() {
    this.setMode(Mode.Edit);
  }

  /**
   * @return deselects the element
   */
  public void deSelect() {
    this.setMode(Mode.Normal);
  }

  /**
   * @return the mode
   */
  public Mode getMode() {
    return mode;
  }

  /**
   * @param mode the mode to set
   */
  public void setMode(Mode mode) {
    switch(this.getMode()) {
      case Normal:
        switch(mode) {
          case Normal:
            break;
          case Edit:
            this.calculateEditFromNormal();
            break;
          case TextEdit:
            this.calculateEditFromNormal();
            this.calculateTextFromEdit();
            break;
          default:
            System.err.println("ERROR occurred: Mode-Switch from Normal to " + mode + " is not allowed!");
            break;
        }
        break;
        
      case InitialCreate:
        switch(mode) {
          case Normal:
            this.calculateEditFromCreate();
            this.calculateNormalFromEdit();
            break;
          case Edit:
            this.calculateEditFromCreate();
            break;
          case ExtendedCreate:
            this.mode = mode;
            break;
          default:
            System.err.println("ERROR occurred: Mode-Switch from InitialCreate to " + mode + " is not allowed!");
            break;
        }
        break;
      case ExtendedCreate:
        switch(mode) {
          case Edit:
            this.calculateEditFromCreate();
            break;
          default:
            System.err.println("ERROR occurred: Mode-Switch from ExtendedCreate to " + mode + " is not allowed!");
            break;
        }
        break;
        
      case Edit:
        switch(mode) {
          case Normal:
            this.calculateNormalFromEdit();
            break;
          case Edit:
            break;
          case ExtendedEdit:
            this.calculateExtendedFromEdit();
            break;
          case TextEdit:
            this.calculateTextFromEdit();
            break;
          default:
            System.err.println("ERROR occurred: Mode-Switch from Edit to " + mode + " is not allowed!");
            break;
        }
        break;
        
      case ExtendedEdit:
        switch(mode) {
          case Normal:
            this.calculateEditFromExtended();
            this.calculateNormalFromEdit();
            break;
          case Edit:
            this.calculateEditFromExtended();
            break;
          case ExtendedEdit:
            break;
          default:
            System.err.println("ERROR occurred: Mode-Switch from ExtendedEdit to " + mode + " is not allowed!");
            break;
        }
        break;
        
      case TextEdit:
        if(!this.getClass().isAssignableFrom(SVGText.class)) {
          String text = this.getText();
          while(text.endsWith("\n") || text.endsWith("\r")) {
            text = text.substring(0, text.length()-1);
          }
          this.setText(text);
        }
        switch(mode) {
          case Normal:
            this.calculateEditFromText();
            this.calculateNormalFromEdit();
            ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(ActionPerformedEvent.Action.createHistoryPoint));
            break;
          case Edit:
            this.calculateEditFromText();
            ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(ActionPerformedEvent.Action.createHistoryPoint));
            break;
          default:
            System.err.println("ERROR occurred: Mode-Switch from TextEdit to " + mode + " is not allowed!");
            break;
        }
        break;
    }
  }
  
  /**
   * @return if element is in InitialCreate Mode
   */
  public boolean isInitialCreateMode() {
    return (this.getMode()==Mode.InitialCreate);
  }
  
  /**
   * @return if element is in ExtendedCreate Mode
   */
  public boolean isExtendedCreateMode() {
    return (this.getMode()==Mode.ExtendedCreate);
  }
  
  /**
   * @return if element is in Edit Mode
   */
  public boolean isEditMode() {
    return (this.getMode()==Mode.Edit);
  }
  
  /**
   * @return if element is in ExtendedEdit Mode
   */
  public boolean isExtendedEditMode() {
    return (this.getMode()==Mode.ExtendedEdit);
  }
  
  /**
   * @return if element is in TextEdit Mode
   */
  public boolean isTextEditMode() {
    return (this.getMode()==Mode.TextEdit);
  }
  
  /**
   * @return if element is in Normal Mode
   */
  public boolean isNormalMode() {
    return (this.getMode()==Mode.Normal);
  }

  /**
   * @return if element is selected
   */
  public boolean isSelected() {
    return !this.isNormalMode();
  }
  
  /**
   * Gets the Cursor for the Normal-Mouse-Over
   * @return Cursor to be set.
   */
  public Cursor getNormalCursor() {
    return new Cursor(Cursor.HAND_CURSOR);
  }
  
  /**
   * Gets the Cursor for the Edit-Mode
   * @return Cursor to be set.
   */
  public Cursor getEditCursor() {
    return new Cursor(Cursor.MOVE_CURSOR);
  }
  
  /**
   * Gets the Cursor for the ExtendedEdit-Mode
   * @return Cursor to be set.
   */
  public Cursor getExtendedEditCursor() {
    return new Cursor(Cursor.MOVE_CURSOR);
  }
  
  /**
   * Gets the Cursor for the InitialCreate-Mode
   * @return Cursor to be set.
   */
  public Cursor getInitialCreateCursor() {
    BufferedImage buffer = new BufferedImage(32, 32, BufferedImage.TYPE_INT_ARGB);
    Graphics2D g = (Graphics2D) buffer.getGraphics();
    g.setStroke(new BasicStroke(1, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER));
    g.setRenderingHint(RenderingHints.KEY_RENDERING,
                       RenderingHints.VALUE_RENDER_QUALITY);
    g.setRenderingHint(RenderingHints.KEY_ALPHA_INTERPOLATION,
                       RenderingHints.VALUE_ALPHA_INTERPOLATION_QUALITY);
    //g.transform(new AffineTransform(1.5, 0, 0, 1.5, 0, 0));
    g.setColor(Color.BLACK);
    g.drawLine(0, 7, 14, 7);
    g.drawLine(7, 0, 7, 14);

    return Toolkit.getDefaultToolkit().createCustomCursor(buffer, new Point(7,7), "InitialCreate Cursor");
  }
  
  /**
   * Gets the Cursor for the ExtendedCreate-Mode
   * @return Cursor to be set.
   */
  public Cursor getExtendedCreateCursor() {
    //return new Cursor(Cursor.CROSSHAIR_CURSOR);
    BufferedImage buffer = new BufferedImage(32, 32, BufferedImage.TYPE_INT_ARGB);
    Graphics2D g = (Graphics2D) buffer.getGraphics();
    g.setStroke(new BasicStroke(1, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER));
    g.setRenderingHint(RenderingHints.KEY_RENDERING,
                       RenderingHints.VALUE_RENDER_QUALITY);
    g.setRenderingHint(RenderingHints.KEY_ALPHA_INTERPOLATION,
                       RenderingHints.VALUE_ALPHA_INTERPOLATION_QUALITY);
    //g.transform(new AffineTransform(1.5, 0, 0, 1.5, 0, 0));
    g.setColor(Color.BLACK);
    g.drawLine(0, 7, 14, 7);
    g.drawLine(7, 0, 7, 14);

    return Toolkit.getDefaultToolkit().createCustomCursor(buffer, new Point(7,7), "ExtendedCreate Cursor");
  }
  
  /**
   * Gets the Cursor for Mouse-Over-Handle.
   * @return Cursor to be set.
   */
  public Cursor getHandleCursor() {
    //return new Cursor(Cursor.CROSSHAIR_CURSOR);
    BufferedImage buffer = new BufferedImage(32, 32, BufferedImage.TYPE_INT_ARGB);
    Graphics2D g = (Graphics2D) buffer.getGraphics();
    g.setStroke(new BasicStroke(1.5f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER));
    g.setRenderingHint(RenderingHints.KEY_RENDERING,
                       RenderingHints.VALUE_RENDER_QUALITY);
    g.setRenderingHint(RenderingHints.KEY_ALPHA_INTERPOLATION,
                       RenderingHints.VALUE_ALPHA_INTERPOLATION_QUALITY);
    //g.transform(new AffineTransform(1.5, 0, 0, 1.5, 0, 0));
    g.setColor(Color.BLACK);
    g.drawLine(0, 7, 14, 7);
    g.drawLine(7, 0, 7, 14);

    return Toolkit.getDefaultToolkit().createCustomCursor(buffer, new Point(7,7), "Handle Cursor");
  }
  
  /**
   * Gets the Cursor for TextEdit Mode
   * @return Cursor to be set.
   */
  public Cursor getTextEditCursor() {
    return new Cursor(Cursor.TEXT_CURSOR);
  }
  
  
  /**
   * This Method calculates the Edit-Mode from Normal-Mode
   */
  protected void calculateEditFromNormal() {
    this.editHandles.clear();
    for(Point p : this.normalHandles) {
      this.editHandles.add((Point)p.clone());
    }
    this.updateBoundingRect(this.editHandles);
    this.mode = Mode.Edit;
  }

  /**
   * This Method calculates the Edit-Mode from Initial-/Extended-Create-Mode
   */
  protected void calculateEditFromCreate() {
    this.editHandles.clear();
    for(Point p : this.initHandles) {
      this.editHandles.add((Point)p.clone());
    }
    this.updateBoundingRect(this.editHandles);
    this.mode = Mode.Edit;
  }

  /**
   * This method calculates the Normal Mode from Edit Mode.
   */
  protected void calculateNormalFromEdit() {
    if(this.validate()) {
      this.normalHandles.clear();
      this.normalHandles.addAll((Vector<Point>)this.editHandles.clone());
    }
    this.editHandles.clear();
    this.updateBoundingRect(this.normalHandles);
    this.mode = Mode.Normal;
  }
  
  /**
   * This method calculates the Edit Mode from Text Mode.
   */
  protected void calculateEditFromText() {
    this.mode = Mode.Edit;
    KeyInputManager.removeKeyInputReceiver(this);
  }
  
  /**
   * This method calculates the Normal Mode from Text Mode.
   */
  protected void calculateNormalFromText() {
    this.mode = Mode.Normal;
    KeyInputManager.removeKeyInputReceiver(this);
  }
  
  /**
   * This method calculates the Text Mode from Edit Mode.
   */
  protected void calculateTextFromEdit() {
    this.mode = Mode.TextEdit;
    KeyInputManager.registerKeyInputReceiver(this);
  }
  
  /**
   * This method calculates the Extended Mode from Edit Mode.
   */
  protected void calculateExtendedFromEdit() { 
    this.extendedHandles.clear();
    for(Point p : this.editHandles) {
      this.extendedHandles.add((Point)p.clone());
    }
    this.mode = Mode.ExtendedEdit;
  }

  /**
   * This method calculates the Edit Mode from Extended Mode.
   */
  protected void calculateEditFromExtended() { 
    if(this.validate()) {
      this.editHandles.clear();
      this.editHandles.addAll((Vector<Point>)this.extendedHandles.clone());
    }
    this.extendedHandles.clear();
    this.updateBoundingRect(this.editHandles);
    this.mode = Mode.Edit;
  }


 /*----------------------------------------------------------------------------
  * 
  * Getter / Setter
  * 
  ----------------------------------------------------------------------------*/
  
  
    /**
   * @return the graphics
   */
  public Graphics2D getGraphics() {
    return graphics;
  }

  /**
   * @param graphics the graphics to set
   */
  public void setGraphics(Graphics2D graphics) {
    this.graphics = (Graphics2D)graphics.create();
    this.graphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                                   RenderingHints.VALUE_ANTIALIAS_ON);
    this.graphics.setRenderingHint(RenderingHints.KEY_RENDERING, 
                                   RenderingHints.VALUE_RENDER_QUALITY);
    this.graphics.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
    this.graphics.setRenderingHint(RenderingHints.KEY_STROKE_CONTROL, RenderingHints.VALUE_STROKE_NORMALIZE);
  }
  
  /**
   * @param fillColor the fillColor to set
   */
  public void setFillColor(Color fillColor) {
    this.fillColor = fillColor;
  }

  /**
   * @return the fillColor
   */
  public Color getFillColor() {
    return fillColor;
  }

  /**
   * @return the lineColor
   */
  public Color getLineColor() {
    return lineColor;
  }

  /**
   * @param lineColor the lineColor to set
   */
  public void setLineColor(Color lineColor) {
    this.lineColor = lineColor;
  }

  /**
   * @return the text
   */
  public String getText() {
    return text;
  }

  /**
   * @param text the text to set
   */
  public void setText(String text) {
    if(this.isTextSetable())
      this.text = text;
  }

  /**
   * @return the textSize
   */
  public int getTextSize() {
    return textSize;
  }

  /**
   * @param textSize the textSize to set
   */
  public void setTextSize(int textSize) {
    this.textSize = textSize;
  }

  /**
   * @return the lineStrength
   */
  public int getLineStrength() {
    return lineStrength;
  }
  
  /**
   * @return the lineStrength
   */
  public double getExportLineStrength() {
    if (this.getLineStrength() > 0) {
      return this.getLineStrength();
    } else if (this.lineStrength == 0) {// Hairline
      return 0.25;
    }
    // No Line
    if(!this.isFillcolorSetable())
      return 1.0; // If not Fillable, line is obligated
    return 0.0;

  }

  /**
   * @param lineStrength the lineStrength to set
   */
  public void setLineStrength(int lineStrength) {
    this.lineStrength = lineStrength;
  }

  /**
   * @return the fillcolorSetable
   */
  public boolean isFillcolorSetable() {
    return fillcolorSetable;
  }

  /**
   * @param fillcolorSetable the fillcolorSetable to set
   */
  public void setFillcolorSetable(boolean fillcolorSetable) {
    this.fillcolorSetable = fillcolorSetable;
  }

  /**
   * @return the textSetable
   */
  public boolean isTextSetable() {
    return textSetable;
  }

  /**
   * @param textSetable the textSetable to set
   */
  public void setTextSetable(boolean textSetable) {
    this.textSetable = textSetable;
  }
  
    /**
   * @return the font
   */
  public Font getFont() {
    return font;
  }

  /**
   * @param font the font to set
   */
  public void setFont(Font font) {
    this.font = font;
  }

  /**
   * @return the textColor
   */
  public Color getTextColor() {
    return textColor;
  }

  /**
   * @param textColor the textColor to set
   */
  public void setTextColor(Color textColor) {
    this.textColor = textColor;
  }

  /**
   * @return the textAlignment
   */
  public Alignment getTextAlignment() {
    return textAlignment;
  }

  /**
   * @param textAlignment the textAlignment to set
   */
  public void setTextAlignment(Alignment textAlignment) {
    this.textAlignment = textAlignment;
  }

  /**
   * @return the textOption
   */
  public ArrayList<TextOption> getTextOption() {
    return textOption;
  }

  /**
   * @param textOption the textOption to set
   */
  public void setTextOption(ArrayList<TextOption> textOption) {
    this.textOption = (ArrayList<TextOption>)textOption.clone();
  }

  /**
   * @param textOption the textOption to add
   */
  public void addTextOption(TextOption textOption) {
    if (!this.textOption.contains(textOption)) {
      this.textOption.add(textOption);
    }
  }


  /**
   * Gets the bounding box.
   *
   * @return Rectangle created from the bounding box.
   */
  public Rectangle getBoundingBox() {
    Rectangle r = (this.normalBoundingBox == null) 
                      ? new Rectangle(0, 0, 1, 1) 
                      : this.normalBoundingBox;
    if (r.height == 0)  r.height = 1;
    if (r.width == 0)   r.width = 1;
    return r;
  }

  /**
   * @param boundingRect the boundingRect to set
   */
  public void setBoundingBox(Rectangle boundingRect) {
    this.normalBoundingBox = boundingRect;
  }

  /**
   * @return the textBoundingBox
   */
  public Rectangle getTextBoundingBox() {
    return this.textBoundingBox;
  }

  /**
   * @param textBoundingBox the textBoundinBox to set
   */
  public void setTextBoundingBox(Rectangle textBounds) {
    this.textBoundingBox = textBounds;
  }

  /**
   * @return the textMargin
   */
  public int getTextMargin() {
    return (Math.abs(this.getLineStrength()) + this.standardTextMargin);
  }

  /**
   * @param textMargin the textMargin to set
   
  public void setTextMargin(int textMargin) {
    this.textMargin = textMargin;
  }*/
  
}
