 /**----------------------------------------------------
 * File:      SVGEllipse.java
 * Package:   at.hoertler.svgeditor.svg.elements
 * Project:   SVGEditor
 * ----------------------------------------------------
 * @author Philipp Hörtler BSc
 * @version 1.0.0
 * Date: 2013-03-24
 * ----------------------------------------------------
 * Description:
 * 
 * Represents an SVGElement.
 */
package at.hoertler.svgeditor.svg.elements;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.event.MouseEvent;
import java.awt.geom.Ellipse2D;
import java.util.Vector;
import javax.swing.Icon;

public class SVGEllipse
        extends SVGElement {

  public SVGEllipse() {
    this.setFillcolorSetable(true);
    this.setTextSetable(true);
    this.getHandles().add(new Point(-1,-1));
    this.getHandles().add(new Point(-1,-1));
    this.getHandles().add(new Point(-1,-1));
    this.getHandles().add(new Point(-1,-1));
  }

  @Override
  public SVGElement getInstance() {
    return new SVGEllipse();
  }

  @Override
  public void paint(Graphics g) {
    this.paint(g, 1);
  }

  @Override
  public void paint(Graphics g, int zoomFactor) {
    Vector<Point> handles = this.getHandles();

    this.setGraphics((Graphics2D) g);
    Ellipse2D r = this.createOval(zoomFactor);
    this.getGraphics().setStroke(getStroke(zoomFactor));
    if (this.isFillcolorSetable() && this.getFillColor() != null) {
      this.getGraphics().setColor(this.getFillColor());
      this.getGraphics().fill(r);
    }
    this.getGraphics().setColor(this.getLineColor());
    this.getGraphics().draw(r);


    this.renderText(zoomFactor);

  }

  @Override
  public Icon getIcon() {
    return new Icon() {
      public int getIconHeight() {
        return 16;
      }

      public int getIconWidth() {
        return 16;
      }

      @Override
      public void paintIcon(Component c, Graphics g, int x, int y) {
        Graphics2D g2 = (Graphics2D) g;
        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                            RenderingHints.VALUE_ANTIALIAS_ON);
        g2.setColor(Color.BLACK);
        g2.setStroke(new BasicStroke(1, BasicStroke.CAP_BUTT,
                                     BasicStroke.JOIN_MITER));
        g2.drawOval(x + 1, y + 4, 12, 8);
      }
    };
  }

  @Override
  public void moveHandle(int handleIndex, Point p, MouseEvent e) {
    if (this.isNormalMode())
      return;
    this.getHandles().get(handleIndex).translate(p.x, p.y);
    recalculateHandles(handleIndex, p);
    this.updateBoundingRect(this.getHandles());
  }

  @Override
  public boolean validate() {
    if(this.getHandles().isEmpty()) return false;
    this.reorderHandles();
    if (this.getHandles().get(0).equals(this.getHandles().get(1)) || this.getHandles().get(0).equals(this.getHandles().get(3)))
      return false;
    return true;
  }

  @Override
  public String exportToSVG() {
    Point p1 = this.getHandles().get(0);
    Point p2 = this.getHandles().get(2);
    Ellipse2D r = this.createOval(1);
    StringBuilder sb = new StringBuilder();
    Color c = this.getLineColor();
    sb.append("\t");
    sb.append("<g id=\"" + this.getClass().getSimpleName() + "\">\n");
    sb.append(this.exportTextToSVG());
    sb.append("\t\t");
    sb.append("<ellipse ");
    sb.
            append(
            "cx=\"" + (int)(r.
                       getX() + r.
                       getWidth() / 2) + "\" cy=\"" + (int)(r.
                                                       getY() + r.
                                                       getHeight() / 2) + "\" ");
    sb.append(
            "rx=\"" + (int)r.getWidth() / 2 + "\" ry=\"" + (int)r.getHeight() / 2 + "\" ");
    sb.append("stroke=\"rgb(" + c.getRed() + "," + c.getGreen() + "," + c.
            getBlue() + ")\" ");
    c = this.getFillColor();
    if (c.getAlpha() != 0)
      sb.append("fill=\"rgb(" + c.getRed() + "," + c.getGreen() + "," + c.
              getBlue() + ")\" ");
    else
      sb.append("fill=\"none\" ");
    sb.append("stroke-width=\"" + this.getExportLineStrength() + "\" ");
    sb.append("/>\n");
    sb.append("\t");
    sb.append("</g>\n");
    return sb.toString();
  }

  @Override
  public void handleMousePressed(MouseEvent e) {
    if(!this.isInitialCreateMode()) return;
    
    this.getHandles().set(0, e.getPoint());
    this.getHandles().set(1, e.getPoint());
    this.getHandles().set(2, e.getPoint());
    this.getHandles().set(3, e.getPoint());
    this.updateBoundingRect(this.getHandles());
  }

  @Override
  public void handleMouseDragged(MouseEvent e) {
    if(!this.isInitialCreateMode()) return;
    
    this.getHandles().set(2, e.getPoint());
    this.updateBoundingRect(this.getHandles());
  }

  @Override
  public void handleMouseReleased(MouseEvent e) {
    if(!this.isInitialCreateMode()) return;
    
    reorderHandles();
    this.updateBoundingRect(this.getHandles());
    this.setMode(Mode.Edit);
  }

  @Override
  public void handleMouseClicked(MouseEvent e) {
    if(!this.isEditMode() && !this.isTextEditMode()) return;

    if(e.getClickCount()>=2) {
      if(this.isTextEditMode())
        this.setMode(Mode.Edit);
      else
        this.setMode(Mode.TextEdit);
    }
  }

  @Override
  public boolean hit(Rectangle rect) {
    return createOval(1).intersects(rect);
  }

  private Ellipse2D createOval(int zoomFactor) {
    Point xy = new Point(minX(this.getHandles(), zoomFactor), minY(this.getHandles(), zoomFactor));
    Dimension d = new Dimension(maxX(this.getHandles(), zoomFactor) - xy.x,
                                maxY(this.getHandles(), zoomFactor) - xy.y);
    return new Ellipse2D.Double(xy.x, xy.y, d.width, d.height);
  }

  private void recalculateHandles(int index, Point translation) {
    Vector<Point> handles = this.getHandles();
    switch (index) {
      case 0:
        if (translation.x != 0)
          handles.get(3).translate(translation.x, 0);
        if (translation.y != 0)
          handles.get(1).translate(0, translation.y);
        break;
      case 1:
        if (translation.x != 0)
          handles.get(2).translate(translation.x, 0);
        if (translation.y != 0)
          handles.get(0).translate(0, translation.y);
        break;
      case 2:
        if (translation.x != 0)
          handles.get(1).translate(translation.x, 0);
        if (translation.y != 0)
          handles.get(3).translate(0, translation.y);
        break;
      case 3:
        if (translation.x != 0)
          handles.get(0).translate(translation.x, 0);
        if (translation.y != 0)
          handles.get(2).translate(0, translation.y);
        break;
    }
  }

  private void reorderHandles() {
    Point h0 = new Point(minX(this.getHandles(), 1), minY(this.getHandles(), 1));
    Point h1 = new Point(maxX(this.getHandles(), 1), minY(this.getHandles(), 1));
    Point h2 = new Point(maxX(this.getHandles(), 1), maxY(this.getHandles(), 1));
    Point h3 = new Point(minX(this.getHandles(), 1), maxY(this.getHandles(), 1));
    this.getHandles().get(0).setLocation(h0);
    this.getHandles().get(1).setLocation(h1);
    this.getHandles().get(2).setLocation(h2);
    this.getHandles().get(3).setLocation(h3);
  }

  @Override
  public Point getTranslationPointForHandle(int handleIndex) {
    return this.getHandles().get((handleIndex+2)%4);
  }
}
