 /**----------------------------------------------------
 * File:      SVGLine.java
 * Package:   at.hoertler.svgeditor.svg.elements
 * Project:   SVGEditor
 * ----------------------------------------------------
 * @author Philipp Hörtler BSc
 * @version 1.0.0
 * Date: 2013-03-24
 * ----------------------------------------------------
 * Description:
 * 
 * Represents an SVGElement.
 */
package at.hoertler.svgeditor.svg.elements;

import at.hoertler.svgeditor.common.HelperFunction;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.event.MouseEvent;
import java.awt.geom.Line2D;
import java.awt.geom.Path2D;
import java.util.Vector;
import javax.swing.Icon;

public class SVGLine
        extends SVGElement
        implements SVGArrow {
  private boolean arrowAtFront = false;
  private boolean arrowAtEnd = false;
  
  public SVGLine() {
    this.getHandles().add(new Point(-1,-1));
    this.getHandles().add(new Point(-1,-1));
    this.setTextSetable(true);
  }

  @Override
  public SVGElement getInstance() {
    SVGLine l = new SVGLine();
    l.setHasArrowAtEnd(this.arrowAtEnd);
    l.setHasArrowAtFront(this.arrowAtFront);
    return l;
  }

  @Override
  public void paint(Graphics g) {
    this.paint(g, 1);
  }

  @Override
  public void paint(Graphics g, int zoomFactor) {
    if(this.getHandles().get(0).x == this.getHandles().get(1).x && this.getHandles().get(0).y == this.getHandles().get(1).y)
      return;
    this.setGraphics((Graphics2D) g);

    if(hasArrowAtFront() || hasArrowAtEnd())
      this.paintArrow(zoomFactor);
    else
      this.paintLine(zoomFactor);
   

    this.renderText(zoomFactor);

  }
  
  private void paintLine(int zoomFactor) {
    Vector<Point> handles = this.getHandles();
    Point a = calcZoom(handles.get(0), zoomFactor);
    Point b = calcZoom(handles.get(1), zoomFactor);

    this.getGraphics().setStroke(getStroke(zoomFactor));
    this.getGraphics().setColor(this.getLineColor());
    this.getGraphics().draw(new Line2D.Double(calcZoom(handles.get(0),
                                                       zoomFactor),
                                              calcZoom(handles.get(1),
                                                       zoomFactor)));
  }
  
  private void paintArrow(int zoomFactor) {
    Point a = calcZoom(this.getHandles().get(0), zoomFactor);
    Point b = calcZoom(this.getHandles().get(1), zoomFactor);
    Point start = null, end = null;;
    if(this.hasArrowAtFront()) {
      end = this.paintArrow(a, b, zoomFactor);
    }
    if(this.hasArrowAtEnd()) {
      start = this.paintArrow(b, a, zoomFactor);
    }
    
    if(end==null) end = b;
    if(start==null) start = a;
 
    Line2D l = new Line2D.Double(start, end);
    this.getGraphics().draw(l);
  }
  
  private Point paintArrow(Point a, Point b, int zoomFactor) {
    this.getGraphics().setStroke(getStroke(zoomFactor));
    this.getGraphics().setColor(this.getLineColor());

    Point[] points = calcArrowHead(a, b, zoomFactor);
    
    Point inters = HelperFunction.intersectLines(new Line2D.Float(a,b), new Line2D.Double(points[0],points[1]));

    //Line2D l = new Line2D.Double(a, inters);
    //this.getGraphics().draw(l);
    Path2D.Double p = new Path2D.Double();
    p.moveTo(b.x,b.y);
    p.lineTo(points[0].x,points[0].y);
    p.lineTo(points[1].x,points[1].y);
    p.lineTo(b.x,b.y);
    p.closePath();
    this.getGraphics().fill(p);
    return inters;
  }
  
  private Point[] calcArrowHead(Point a, Point b, int zoomFactor) {
    Point[] points = new Point[2];
    double headLen = this.getArrowHeadLength(zoomFactor);
    double lineLen = Math.sqrt(Math.pow(b.x-a.x, 2) + Math.pow(b.y-a.y, 2));
    double angle0 = Math.tan(15*Math.PI/180);
    double angle1 = angle0*-1;
  
    Point s = new Point((int)(b.x - ((b.x-a.x) * Math.cos(angle0) - Math.sin(angle0) *(b.y-a.y))*headLen/lineLen),
                        (int)(b.y - ((b.y-a.y) * Math.cos(angle0) + Math.sin(angle0) * (b.x-a.x))*headLen/lineLen));
    Point t = new Point((int)(b.x - ((b.x-a.x) * Math.cos(angle1) - Math.sin(angle1) *(b.y-a.y))*headLen/lineLen),
                        (int)(b.y - ((b.y-a.y) * Math.cos(angle1) + Math.sin(angle1) * (b.x-a.x))*headLen/lineLen));
    points[0] = s;
    points[1] = t;
    return points;
  }
  
  private double getArrowHeadLength(int zoomFactor) {
    double ls = this.getLineStrength();
    if(ls>0) ls *= 9.5 ;
    else ls = 5.5;  
    return ls*zoomFactor;
  }

  @Override
  public Icon getIcon() {
    return new Icon() {
      public int getIconHeight() {
        return 16;
      }

      public int getIconWidth() {
        return 16;
      }

      @Override
      public void paintIcon(Component c, Graphics g, int x, int y) {
        Graphics2D g2 = (Graphics2D) g;
        /*g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                            RenderingHints.VALUE_ANTIALIAS_ON);*/
        g2.setColor(Color.BLACK);
        g2.setStroke(new BasicStroke(1, BasicStroke.CAP_BUTT,
                                     BasicStroke.JOIN_MITER));
        g2.drawLine(x + 1, y + 1, x + 14, y + 14);
        
        if(hasArrowAtFront() || hasArrowAtEnd()) {
          g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                            RenderingHints.VALUE_ANTIALIAS_ON);
          Polygon arrowHead = new Polygon();
          arrowHead.addPoint(x, y);
          arrowHead.addPoint(x + 5, y + 9);
          arrowHead.addPoint(x + 9, y + 5);
          g2.fillPolygon(arrowHead);
        }
      }
    };
  }

  @Override
  public void moveHandle(int handleIndex, Point p, MouseEvent e) {
    if (this.isNormalMode()) {
      return;
    }
    this.getHandles().get(handleIndex).translate(p.x, p.y);
    this.updateBoundingRect(this.getHandles());
  }

  @Override
  public boolean validate() {
    if (this.getHandles().isEmpty() || this.getHandles().get(0).equals(this.getHandles().get(1))) {
      return false;
    }

    return true;
  }

  @Override
  public String exportToSVG() {
    Point a = calcZoom(this.getHandles().get(0), 1);
    Point b = calcZoom(this.getHandles().get(1), 1);
    Point[] startArrow = null, endArrow = null;
    Point start = null, end = null;
    if(this.hasArrowAtFront()) {
      endArrow = this.calcArrowHead(a, b, 1);
      end = HelperFunction.intersectLines(new Line2D.Float(a,b), new Line2D.Float(endArrow[0],endArrow[1]));
    }
    if(this.hasArrowAtEnd()) {
      startArrow = this.calcArrowHead(b, a, 1);
      start = HelperFunction.intersectLines(new Line2D.Float(a,b), new Line2D.Float(startArrow[0],startArrow[1]));
    }
    
    
    if(end==null) end = b;
    if(start==null) start = a;
    
    
    Point p1 = this.getHandles().get(0);
    Point p2 = this.getHandles().get(1);

    StringBuilder sb = new StringBuilder();
    Color c = this.getLineColor();
    sb.append("\t");
    sb.append("<g id=\"" + this.getClass().getSimpleName() + "\">\n");
    sb.append(this.exportTextToSVG());
    sb.append("\t\t");
    sb.append("<line ");
    sb.append("x1=\"" + start.x + "\" y1=\"" + start.y + "\" ");
    sb.append("x2=\"" + end.x + "\" y2=\"" + end.y + "\" ");
    sb.append("stroke=\"rgb(" + c.getRed() + "," + c.getGreen() + "," + c.
            getBlue() + ")\" ");
    sb.append("stroke-width=\"" + this.getExportLineStrength() + "\"");
    sb.append(" />");
    sb.append("\n");
    if(this.hasArrowAtEnd()) {
      sb.append("\t\t");
    sb.append("<path ");
    sb.append("d=\"M" + a.x + "," + a.y + 
                 "L" + startArrow[0].x + "," + startArrow[0].y + 
                 "L" + startArrow[1].x + "," + startArrow[1].y + 
                 "L" + a.x + "," + a.y + "\" ");
    
    sb.append("stroke=\"rgb(" + c.getRed() + "," + c.getGreen() + "," + c.
            getBlue() + ")\" ");
    sb.append("fill=\"rgb(" + c.getRed() + "," + c.getGreen() + "," + c.
            getBlue() + ")\" ");
    sb.append("stroke-width=\"0.25\"");
    sb.append(" />");
    sb.append("\n");
  }
  if(this.hasArrowAtFront()) {
    sb.append("\t\t");
    sb.append("<path ");
    sb.append("d=\"M" + b.x + "," + b.y + 
                 "L" + endArrow[0].x + "," + endArrow[0].y + 
                 "L" + endArrow[1].x + "," + endArrow[1].y + 
                 "L" + b.x + "," + b.y + "\" ");
    sb.append("stroke=\"rgb(" + c.getRed() + "," + c.getGreen() + "," + c.
            getBlue() + ")\" ");
    sb.append("fill=\"rgb(" + c.getRed() + "," + c.getGreen() + "," + c.
            getBlue() + ")\" ");
    sb.append("stroke-width=\"0.25\"");
    sb.append(" />");
    sb.append("\n");
    
  }
  sb.append("\t");
  sb.append("</g>\n");

    return sb.toString();
  }

  @Override
  public void handleMousePressed(MouseEvent e) {
    if(!this.isInitialCreateMode()) return;
    
    this.getHandles().set(0, e.getPoint());
    this.getHandles().set(1, e.getPoint());
    this.updateBoundingRect(this.getHandles());
  }

  @Override
  public void handleMouseDragged(MouseEvent e) {
    if(!this.isInitialCreateMode()) return;
    
    this.getHandles().set(1, e.getPoint());
    this.updateBoundingRect(this.getHandles());
  }

  @Override
  public void handleMouseReleased(MouseEvent e) {
    if(!this.isInitialCreateMode()) return;
    
    this.updateBoundingRect(this.getHandles());
    this.setMode(Mode.Edit);
  }

  @Override
  public void handleMouseClicked(MouseEvent e) {
    if(!this.isEditMode() && !this.isTextEditMode()) return;
    // If handle is double-clicked, toggle an arrow head at this side
    if(e.getClickCount()>=2) {
      int handle = this.getHandleIndex(e.getPoint());
      if(handle != -1) {
        if (handle == 0)
          this.setHasArrowAtEnd(!this.hasArrowAtEnd());
        if (handle == 1)
          this.setHasArrowAtFront(!this.hasArrowAtFront());
      } else {
        if(this.isTextEditMode())
          this.setMode(Mode.Edit);
        else
          this.setMode(Mode.TextEdit);
      }
    }
  }

  @Override
  public boolean hit(Rectangle rect) {
    Vector<Point> handles = this.getHandles();
    Line2D l = new Line2D.Double(handles.get(0), handles.get(1));
    return l.intersects(rect);
  }

  /**
   * @return the hasArrowAtFront
   */
  public boolean hasArrowAtFront() {
    return arrowAtFront;
  }

  /**
   * @param hasArrowAtFront the hasArrowAtFront to set
   */
  public void setHasArrowAtFront(boolean hasArrowAtFront) {
    this.arrowAtFront = hasArrowAtFront;
    if(this.hasArrowAtFront() || this.hasArrowAtEnd())
      this.standardTextMargin = (int)this.getArrowHeadLength(1)+5;
    else
      this.standardTextMargin = 5;
  }

  /**
   * @return the hasArrowAtEnd
   */
  public boolean hasArrowAtEnd() {
    return arrowAtEnd;
  }

  /**
   * @param hasArrowAtEnd the hasArrowAtEnd to set
   */
  public void setHasArrowAtEnd(boolean hasArrowAtEnd) {
    this.arrowAtEnd = hasArrowAtEnd;
    if(this.hasArrowAtFront() || this.hasArrowAtEnd())
      this.standardTextMargin = (int)this.getArrowHeadLength(1)+5;
    else
      this.standardTextMargin = 5;
  }
  
  @Override
  public Point getTranslationPointForHandle(int handleIndex) {
    return this.getHandles().get((handleIndex+1)%2);
  }
  
  @Override
  public void updateTextBoundingBox() {
    this.setTextBoundingBox( new Rectangle(
                          this.getBoundingBox().x + this.getTextMargin(), 
                          this.getBoundingBox().y + this.getTextMargin() - this.getTextSize()/2, 
                          this.getBoundingBox().width - this.getTextMargin()*2, 
                          this.getBoundingBox().height - this.getTextMargin()*2)
                        );
    
  }
}
