 /**----------------------------------------------------
 * File:      SVGPolygone.java
 * Package:   at.hoertler.svgeditor.svg.elements
 * Project:   SVGEditor
 * ----------------------------------------------------
 * @author Philipp Hörtler BSc
 * @version 1.0.0
 * Date: 2013-03-24
 * ----------------------------------------------------
 * Description:
 * 
 * Represents an SVGElement.
 */
package at.hoertler.svgeditor.svg.elements;

import at.hoertler.svgeditor.action.ActionManager;
import at.hoertler.svgeditor.common.HelperFunction;
import at.hoertler.svgeditor.event.ActionPerformedEvent;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.event.MouseEvent;
import java.awt.geom.Line2D;
import java.awt.geom.Path2D;
import java.awt.geom.Rectangle2D;
import java.util.Vector;
import javax.swing.Icon;

public class SVGPolygone
        extends SVGElement
        implements SVGArrow {
  private boolean arrowAtFront = false;
  private boolean arrowAtEnd = false;
  private boolean closedPolygone = false;
  
  public SVGPolygone() {
    this.getHandles().add(new Point(-1,-1));
    this.getHandles().add(new Point(-1,-1));
    this.setFillcolorSetable(true);
    this.setTextSetable(false);
  }

  @Override
  public SVGElement getInstance() {
    SVGPolygone l = new SVGPolygone();
    l.setHasArrowAtEnd(this.arrowAtEnd);
    l.setHasArrowAtFront(this.arrowAtFront);
    return l;
  }

  @Override
  public void paint(Graphics g) {
    this.paint(g, 1);
  }

  @Override
  public void paint(Graphics g, int zoomFactor) {
    if(this.getHandles().get(0).x == this.getHandles().get(1).x 
       && this.getHandles().get(0).y == this.getHandles().get(1).y)
      return;
    this.setGraphics((Graphics2D) g);

    this.paintLine(zoomFactor);
    this.renderText(zoomFactor);
  }
  
  private void paintLine(int zoomFactor) {
    Vector<Point> handles = null;
    if(this.isEditMode())
      handles = this.normalHandles;
    else 
      handles = this.getHandles();
    
    
    Point a = calcZoom(handles.get(0), zoomFactor);
    Point b = calcZoom(handles.get(1), zoomFactor);
    Point c = calcZoom(handles.get(handles.size()-2), zoomFactor);
    Point d = calcZoom(handles.get(handles.size()-1), zoomFactor);
    Point start = a;
    Point end = d;
    if(this.hasArrowAtFront()) {
      Point[] points = calcArrowHead(c, d, zoomFactor);
      end = HelperFunction.intersectLines(new Line2D.Float(c,d), new Line2D.Double(points[0],points[1]));
    }
    if(this.hasArrowAtEnd()) {
      Point[] points = calcArrowHead(b, a, zoomFactor);
      start = HelperFunction.intersectLines(new Line2D.Float(a,b), new Line2D.Double(points[0],points[1]));
    }
    this.getGraphics().setStroke(getStroke(zoomFactor));
    int[] x = new int[handles.size()];
    int[] y = new int[handles.size()];
    x[0] = start.x;
    y[0] = start.y;
    for(int i = 1; i<handles.size()-1; i++) {
      Point tmp = calcZoom(handles.get(i), zoomFactor);
      x[i] = tmp.x;
      y[i] = tmp.y;
    }
    x[handles.size()-1] = end.x;
    y[handles.size()-1] = end.y;
    
    
    if(this.isFillcolorSetable()) {
      this.getGraphics().setColor(this.getFillColor());
      this.getGraphics().fillPolygon(x,y,handles.size());
    }
    if(this.getLineStrength()!=-1) {
      this.getGraphics().setColor(this.getLineColor());
      if(this.isClosedPolygone())
        this.getGraphics().drawPolygon(x,y,handles.size());
      else
        this.getGraphics().drawPolyline(x,y,handles.size());
      if(this.hasArrowAtFront()) {
        this.paintArrow(c, d, zoomFactor);
      }
      if(this.hasArrowAtEnd()) {
        this.paintArrow(b, a, zoomFactor);
      }
    }
  }
   
  private void paintArrow(Point a, Point b, int zoomFactor) {
    this.getGraphics().setStroke(new BasicStroke(1));
    this.getGraphics().setColor(this.getLineColor());

    Point[] points = calcArrowHead(a, b, zoomFactor);
    
    Point inters = HelperFunction.intersectLines(new Line2D.Float(a,b), new Line2D.Double(points[0],points[1]));

    Path2D.Double p = new Path2D.Double();
    p.moveTo(b.x,b.y);
    p.lineTo(points[0].x,points[0].y);
    p.lineTo(points[1].x,points[1].y);
    p.lineTo(b.x,b.y);
    p.closePath();
    this.getGraphics().fill(p);
  }
  
  private Point[] calcArrowHead(Point a, Point b, int zoomFactor) {
    Point[] points = new Point[2];
    double ls = this.getLineStrength();
    if(ls>0) ls *= 9.5 ;
    else ls = 5.5;  
    double headLen = ls*zoomFactor;
    double lineLen = Math.sqrt(Math.pow(b.x-a.x, 2) + Math.pow(b.y-a.y, 2));
    double angle0 = Math.tan(15*Math.PI/180);
    double angle1 = angle0*-1;
  
    Point s = new Point((int)(b.x - ((b.x-a.x) * Math.cos(angle0) - Math.sin(angle0) *(b.y-a.y))*headLen/lineLen),
                        (int)(b.y - ((b.y-a.y) * Math.cos(angle0) + Math.sin(angle0) * (b.x-a.x))*headLen/lineLen));
    Point t = new Point((int)(b.x - ((b.x-a.x) * Math.cos(angle1) - Math.sin(angle1) *(b.y-a.y))*headLen/lineLen),
                        (int)(b.y - ((b.y-a.y) * Math.cos(angle1) + Math.sin(angle1) * (b.x-a.x))*headLen/lineLen));
    points[0] = s;
    points[1] = t;
    return points;
  }

  @Override
  public Icon getIcon() {
    return new Icon() {
      public int getIconHeight() {
        return 16;
      }

      public int getIconWidth() {
        return 16;
      }

      @Override
      public void paintIcon(Component c, Graphics g, int x, int y) {
        Graphics2D g2 = (Graphics2D) g;
        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                            RenderingHints.VALUE_ANTIALIAS_ON);
        g2.setColor(Color.BLACK);
        g2.setStroke(new BasicStroke(1, BasicStroke.CAP_BUTT,
                                     BasicStroke.JOIN_MITER));

        int[] xa = {x + 1,
                    x + 14,
                    x + 14,
                    x + 9,
                    x + 12,
                    x + 5
        };
        int[] ya = {y + 14,
                    y + 14,
                    y + 9,
                    y + 9,
                    y + 1,
                    y + 1
        };
        g2.drawPolygon(xa, ya, 6);

      }
    };
  }

  @Override
  public void moveHandle(int handleIndex, Point p, MouseEvent e) {
    if (this.isNormalMode()) {
      return;
    }
    
    
    
    if(this.isExtendedEditMode()) {
      this.getHandles().get(handleIndex).translate(p.x, p.y);
      this.updateBoundingRect(this.getHandles());
      if(handleIndex%2==0) 
        this.calculateSurroundingHandles(handleIndex);
      else if(handleIndex%2==1 && !e.isControlDown())
        this.calculateLineMovement(handleIndex, p);
    }
    
    if(this.isEditMode()) {
      this.getHandles().get(handleIndex).translate(p.x, p.y);
      int leftH = (handleIndex + 3) % 4;
      int rightH = (handleIndex + 5) % 4;
      Point h1 = this.getHandles().get(leftH);
      h1.x = this.getHandles().get((leftH + 5) % 4).x;
      h1.y = this.getHandles().get((leftH + 3) % 4).y;
      Point h2 = this.getHandles().get(rightH);
      h2.x = this.getHandles().get((rightH + 5) % 4).x;
      h2.y = this.getHandles().get((rightH + 3) % 4).y;

      this.updateBoundingRect(this.getHandles());
      this.recalculateElements();
    }
  }

  @Override
  public boolean validate() {
    if (this.getHandles().get(0).equals(this.getHandles().get(1))) {
      return false;
    }

    return true;
  }

  @Override
  public String exportToSVG() {
    Vector<Point> elems = this.getHandles();
    
    Point a = elems.get(0);
    Point b = elems.get(1);
    Point c = elems.get(elems.size()-2);
    Point d = elems.get(elems.size()-1);
    Point[] startArrow = null, endArrow = null;
    Point start = null, end = null;
    if(this.hasArrowAtFront()) {
      endArrow = this.calcArrowHead(c, d, 1);
      end = HelperFunction.intersectLines(new Line2D.Float(c,d), new Line2D.Float(endArrow[0],endArrow[1]));
    }
    if(this.hasArrowAtEnd()) {
      startArrow = this.calcArrowHead(b, a, 1);
      start = HelperFunction.intersectLines(new Line2D.Float(a,b), new Line2D.Float(startArrow[0],startArrow[1]));
    }
    
    
    if(end==null) end = b;
    if(start==null) start = a;
    
    
    Point p1 = elems.get(0);
    Point p2 = elems.get(1);

    StringBuilder sb = new StringBuilder();
    Color col = this.getLineColor();
    sb.append("\t");
    sb.append("<g id=\"" + this.getClass().getSimpleName() + "\">\n");
    sb.append("\t\t");
    sb.append("<path ");
    sb.append("d=\"M");
    sb.append(elems.firstElement().x + "," + elems.firstElement().y);
    for(int i = 1; i<elems.size(); i++) {
      sb.append("L" + elems.get(i).x + "," + elems.get(i).y);
    }
    if(this.isClosedPolygone())
      sb.append("Z");

    sb.append("\" ");
    sb.append("stroke=\"rgb(" + col.getRed() + "," + col.getGreen() + "," + col.
            getBlue() + ")\" ");
    Color fc = this.getFillColor();
    if (fc.getAlpha() != 0)
      sb.append("fill=\"rgb(" + fc.getRed() + "," + fc.getGreen() + "," + fc.
              getBlue() + ")\" ");
    else
      sb.append("fill=\"none\" ");
    sb.append("stroke-width=\"" + this.getExportLineStrength() + "\"");
    sb.append(" />");
    sb.append("\n");
    if(this.hasArrowAtEnd()) {
      sb.append("\t\t");
    sb.append("<path ");
    sb.append("d=\"M" + a.x + "," + a.y + 
                 "L" + startArrow[0].x + "," + startArrow[0].y + 
                 "L" + startArrow[1].x + "," + startArrow[1].y + 
                 "L" + a.x + "," + a.y + "\" ");
    
    sb.append("stroke=\"rgb(" + col.getRed() + "," + col.getGreen() + "," + col.
            getBlue() + ")\" ");
    sb.append("fill=\"rgb(" + col.getRed() + "," + col.getGreen() + "," + col.
            getBlue() + ")\" ");
    sb.append("stroke-width=\"0.25\"");
    sb.append(" />");
    sb.append("\n");
  }
  if(this.hasArrowAtFront()) {
    sb.append("\t\t");
    sb.append("<path ");
    sb.append("d=\"M" + d.x + "," + d.y + 
                 "L" + endArrow[0].x + "," + endArrow[0].y + 
                 "L" + endArrow[1].x + "," + endArrow[1].y + 
                 "L" + d.x + "," + d.y + "\" ");
    sb.append("stroke=\"rgb(" + col.getRed() + "," + col.getGreen() + "," + col.
            getBlue() + ")\" ");
    sb.append("fill=\"rgb(" + col.getRed() + "," + col.getGreen() + "," + col.
            getBlue() + ")\" ");
    sb.append("stroke-width=\"0.25\"");
    sb.append(" />");
    sb.append("\n");
    
  }
  sb.append("\t");
  sb.append("</g>\n");

    return sb.toString();
  }

  @Override
  public void handleMouseMoved(MouseEvent e) {
    //if(!this.initalEditing && this.isCreateMode())
    if(this.isExtendedCreateMode())
      this.getHandles().lastElement().setLocation(e.getPoint());
  }

  @Override
  public void handleMousePressed(MouseEvent e) {
    if(this.isInitialCreateMode()) {
      this.getHandles().set(0, e.getPoint());
      this.getHandles().set(1, e.getPoint());
      this.updateBoundingRect(this.getHandles());
    }
  }

  @Override
  public void handleMouseDragged(MouseEvent e) {
    if(this.isInitialCreateMode()) {
      this.getHandles().set(1, e.getPoint());
      this.updateBoundingRect(this.getHandles());
    }
  }

  @Override
  public void handleMouseReleased(MouseEvent e) {
    if(this.isInitialCreateMode() && !this.getHandles().get(0).equals(e.getPoint())) {
      this.getHandles().set(1, e.getPoint());
      this.updateBoundingRect(this.getHandles());
      this.getHandles().add(new Point(e.getPoint()));
      this.setMode(Mode.ExtendedCreate);
    }
  }

  @Override
  public void handleMouseClicked(MouseEvent e) {
    int handle = -1;
    // If handle is double-clicked, toggle an arrow head at this side
    if(e.getClickCount()>=2) {
      switch(this.getMode()) {
        case ExtendedCreate:
          this.getHandles().remove(this.getHandles().size()-1);
          this.setMode(Mode.Edit);
          handle = this.getHandleIndex(e.getPoint());
          if(handle == 0) {
            this.setClosedPolygone(true);
            this.getHandles().remove(this.getHandles().size()-1);
          } 
          break;
        case ExtendedEdit:
          //this.removeExtendedHandles();
          this.setMode(Mode.Edit);
          break;
        case Edit:
          if (this.hit(e.getPoint())) {
            this.setMode(Mode.ExtendedEdit);
          }
          break;
      }

    } else {
      switch(this.getMode()) {
        case ExtendedCreate:
          handle = this.getHandleIndex(e.getPoint());
          if(handle == 0) {
            this.setClosedPolygone(true);
            this.getHandles().remove(this.getHandles().size()-1);
            this.setMode(Mode.Edit);
          } else if(handle==this.getHandles().size()-2) {
            this.getHandles().remove(this.getHandles().size()-1);
            this.setMode(Mode.Edit);
          } else {
            this.getHandles().add(new Point(e.getPoint()));
          }
          ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(ActionPerformedEvent.Action.createHistoryPoint));
          break;
        case ExtendedEdit:
          if (this.hit(e.getPoint())) {
            handle = this.getHandleIndex(e.getPoint());
            if(handle != -1 && !this.isClosedPolygone()) {
            if (handle == 0)
              this.setHasArrowAtEnd(!this.hasArrowAtEnd());
            else if (handle == this.getHandles().size()-1)
              this.setHasArrowAtFront(!this.hasArrowAtFront());
            }
            ActionManager.fireActionPerformedEvent(new ActionPerformedEvent(ActionPerformedEvent.Action.createHistoryPoint));
          }
          break;
      }
    }
  }

  @Override
  public boolean hit(Rectangle rect) {
    Vector<Point> handles;
    handles = this.getHandles();
    
    Line2D l = null;
    Polygon poly = new Polygon();
    for(int i = 0; i<handles.size(); i++)
      poly.addPoint(handles.get(i).x, handles.get(i).y);
   return poly.intersects(rect);
  }

  /**
   * @return the hasArrowAtFront
   */
  public boolean hasArrowAtFront() {
    return arrowAtFront;
  }

  /**
   * @param hasArrowAtFront the hasArrowAtFront to set
   */
  public void setHasArrowAtFront(boolean hasArrowAtFront) {
    this.arrowAtFront = hasArrowAtFront;
  }

  /**
   * @return the hasArrowAtEnd
   */
  public boolean hasArrowAtEnd() {
    return arrowAtEnd;
  }

  /**
   * @param hasArrowAtEnd the hasArrowAtEnd to set
   */
  public void setHasArrowAtEnd(boolean hasArrowAtEnd) {
    this.arrowAtEnd = hasArrowAtEnd;
  }
  
  /**
   * This method is used to add handles to the polygon.
   * The importer uses this interface to create an polygone.
   * @param p HandlePoint to add.
   */
  public void addHandle(Point p) {
    this.initHandles.add(p);
  }

  /**
   * @return the closedPolygone
   */
  public boolean isClosedPolygone() {
    return closedPolygone;
  }

  /**
   * @param closedPolygone the closedPolygone to set
   */
  public void setClosedPolygone(boolean closedPolygone) {
    this.closedPolygone = closedPolygone;
  }
  
  @Override
  public Point getTranslationPointForHandle(int handleIndex) {
    int s = this.getHandles().size();
    if(this.isExtendedEditMode()) {
      if(!this.isClosedPolygone() && handleIndex==0)
        return this.getHandles().get(1);
      return this.getHandles().get((handleIndex+s-1)%s);
    } else {
      return this.getHandles().get((handleIndex+2)%4);
    }
  }

    
  @Override
  public void moveHandleFinished(int handleIndex, MouseEvent e) {
    if(!this.isExtendedEditMode() || !e.isControlDown())
      return;
    if(handleIndex%2==1) {
      this.getHandles().add(handleIndex, new Point());
      handleIndex++;
      this.getHandles().add(handleIndex+1, new Point());
    }
    this.calculateSurroundingHandles(handleIndex);
  }
  
  private void insertExtendedHandles() {
    Vector<Point> polyEditHdl = new Vector<Point>();
    for(int i = 0; i < this.getHandles().size()-1; i++) {
      polyEditHdl.add((Point)this.getHandles().get(i).clone());
      int x = this.getHandles().get(i).x + (this.getHandles().get(i+1).x - this.getHandles().get(i).x)/2;
      int y = this.getHandles().get(i).y + (this.getHandles().get(i+1).y - this.getHandles().get(i).y)/2;
      polyEditHdl.add(new Point(x,y));
    }
    polyEditHdl.add((Point)this.getHandles().lastElement().clone());

    if(this.isClosedPolygone()) {
      int x = this.getHandles().lastElement().x + (this.getHandles().firstElement().x - this.getHandles().lastElement().x)/2;
      int y = this.getHandles().lastElement().y + (this.getHandles().firstElement().y - this.getHandles().lastElement().y)/2;
      polyEditHdl.add(new Point(x,y));
    }
    this.setHandles(polyEditHdl);
  }

  
  private void removeExtendedHandles() {
    Vector<Point> polyEditHdl = new Vector<Point>();
    for(int i = 0; i < this.getHandles().size(); i+=2) {
      polyEditHdl.add((Point)this.getHandles().get(i).clone());
    }
    this.setHandles(polyEditHdl);
  }
  
  /**
   * This method calculates the relevant handles in case that a whole line has been moved.
   * @param handleIndex HandleIndex which is in the center of the Line that should be moved.
   * @param p Translationpoint to move the line.
   */
  private void calculateLineMovement(int handleIndex, Point p) {
    int s = this.getHandles().size();
    //  a   b   c   m   d   e   f
    //  O---o---O---o---O---o---O
    //  
    // MovePoint = m
    // move line C-D
    // translate c and d like m
    // calculate b from a-c
    // calculate d from d-f
    Point a = this.getHandles().get((handleIndex+s-3)%s);
    //Point b = this.getHandles().get((handleIndex+s-2)%s);
    Point c = this.getHandles().get((handleIndex+s-1)%s);
    Point m = this.getHandles().get(handleIndex);
    Point d = this.getHandles().get((handleIndex+s+1)%s);
    //Point e = this.getHandles().get((handleIndex+s+2)%s);
    Point f = this.getHandles().get((handleIndex+s+3)%s);
    
    // Move Line C-D
    c.translate(p.x, p.y);
    d.translate(p.x, p.y);
    
    if(this.isClosedPolygone() || handleIndex >=3) {
      // Set b
      int x = a.x + (c.x - a.x)/2;
      int y = a.y + (c.y - a.y)/2;
      this.getHandles().get((handleIndex+s-2)%s).setLocation(x, y);
    }
    if(this.isClosedPolygone() || handleIndex <= s-3) {
      // Set e
      int x = f.x + (d.x - f.x)/2;
      int y = f.y + (d.y - f.y)/2;
      this.getHandles().get((handleIndex+s+2)%s).setLocation(x, y);
    }
  }
  
  /**
   * This method calculates the surrounding additional Handles that are inserted when in Edit-Mode.
   * @param handleIndex The handleIndex of the Handle that has been moved.
   */
  private void calculateSurroundingHandles(int handleIndex) {
    int s = this.getHandles().size();
    //  a   b   m    c   d
    //  O---o---O---o---O
    //
    // calculate b from a-m
    // calculate c from d-m

    Point a = this.getHandles().get((handleIndex+s-2)%s);
    //Point b = this.getHandles().get((handleIndex+s-1)%s);
    Point m = this.getHandles().get(handleIndex);
    //Point c = this.getHandles().get((handleIndex+s+1)%s);
    Point d = this.getHandles().get((handleIndex+s+2)%s);

    if(this.isClosedPolygone() || handleIndex >=2) {
      // Set b
      int x = a.x + (m.x - a.x)/2;
      int y = a.y + (m.y - a.y)/2;
      this.getHandles().get((handleIndex+s-1)%s).setLocation(x, y);
    }
    if(this.isClosedPolygone() || handleIndex <= s-2) {
      // Set c
      int x = d.x + (m.x - d.x)/2;
      int y = d.y + (m.y - d.y)/2;
      this.getHandles().get((handleIndex+s+1)%s).setLocation(x, y);
    }
  }
  
  @Override
  public void paintHandle(Graphics g, int handleIndex, int zoomFactor) {
    if(this.isExtendedEditMode() && handleIndex%2==1) {
      ((Graphics2D) g).setStroke(new BasicStroke(1, BasicStroke.CAP_BUTT,
                                                 BasicStroke.JOIN_MITER));
      Rectangle2D r = new Rectangle2D.Double();
      Point hdl = this.getHandles().get(handleIndex);
      Point p = new Point(hdl.x * zoomFactor, hdl.y * zoomFactor);
      r.setRect(p.x - 2, p.y - 2, 4, 4);
      g.setColor(Color.BLACK);
      ((Graphics2D) g).fill(r);
      g.setColor(Color.WHITE);
      ((Graphics2D) g).draw(r);
    } else {
      super.paintHandle(g, handleIndex, zoomFactor);
    }
  }
  
  @Override
  public SVGElement clone() {
    SVGPolygone poly = (SVGPolygone) super.clone();
    poly.closedPolygone = this.closedPolygone;
    return poly;
  }
  
  @Override
  public void translate(int x, int y) {
    super.translate(x, y);
    if (!this.isEditMode())
      return;
    
    /*for(Point p : this.normalHandles)
      p.translate(x, y);*/
    
    this.updateBoundingRect(this.getHandles());
    this.recalculateElements();
  }
  
  @Override
  protected void calculateExtendedFromEdit() { 
    this.extendedHandles.clear();
    for(Point p : this.normalHandles) {
      this.extendedHandles.add((Point)p.clone());
    }
    this.mode = Mode.ExtendedEdit;
    this.insertExtendedHandles();
  }
  
  @Override
  protected void calculateEditFromExtended() { 
    if(!this.validate()) return;
    this.normalHandles.clear();
    this.normalHandles.addAll((Vector<Point>)this.extendedHandles.clone());
    this.updateBoundingRect(this.normalHandles);
    this.extendedHandles.clear();
    
    this.mode = Mode.Normal;
    this.removeExtendedHandles();
    this.mode = Mode.Edit;
    
    this.initHandles.clear();
    for(Point p : this.normalHandles)
      this.initHandles.add((Point)p.clone());
    this.initBoundingBox = (Rectangle)this.getBoundingBox().clone();

    this.getHandles().clear();
    Point p = this.getBoundingBox().getLocation();
    Dimension d = this.getBoundingBox().getSize();
    this.getHandles().add(p);
    this.getHandles().add(new Point(p.x+d.width, p.y));
    this.getHandles().add(new Point(p.x+d.width, p.y+d.height));
    this.getHandles().add(new Point(p.x, p.y+d.height));
  }
  
  @Override
  protected void calculateEditFromNormal() {
    super.calculateEditFromNormal();
    Point p = this.getBoundingBox().getLocation();
    Dimension d = this.getBoundingBox().getSize();
    this.getHandles().clear();
    this.getHandles().add(p);
    this.getHandles().add(new Point(p.x+d.width, p.y));
    this.getHandles().add(new Point(p.x+d.width, p.y+d.height));
    this.getHandles().add(new Point(p.x, p.y+d.height));
  }

  @Override
  protected void calculateEditFromCreate() {
    super.calculateEditFromCreate();
    Point p = this.getBoundingBox().getLocation();
    Dimension d = this.getBoundingBox().getSize();
    super.calculateNormalFromEdit();
    super.calculateEditFromNormal();
    /*this.getHandles().clear();
    this.getHandles().add(p);
    this.getHandles().add(new Point(p.x+d.width, p.y));
    this.getHandles().add(new Point(p.x+d.width, p.y+d.height));
    this.getHandles().add(new Point(p.x, p.y+d.height));*/
  }

  @Override
  protected void calculateNormalFromEdit() {
    if(!this.validate())  return;

    this.initHandles.clear();
    for(Point p : this.normalHandles)
      this.initHandles.add((Point)p.clone());
    this.initBoundingBox = (Rectangle)this.getBoundingBox().clone();
    this.mode = Mode.Normal;
  }
  
  private void recalculateElements() {
    Rectangle oBB = this.initBoundingBox;
    Rectangle aBB = this.getBoundingBox();
    Point transBounding = HelperFunction.
            calculateTranslationPoint(oBB.getLocation(),
                                      aBB.getLocation());
    double scaleW = aBB.width/(double)oBB.width;
    double scaleH = aBB.height/(double)oBB.height;
    
    this.normalHandles.clear();
    // Verzerrung
    for(Point p : this.initHandles) {
      int x = (int)((p.x - oBB.x)*scaleW);
      int y = (int)((p.y - oBB.y)*scaleH);
      Point p1 = new Point(x,y);
      p1.translate(oBB.x, oBB.y);
      p1.translate(transBounding.x, transBounding.y);
      this.normalHandles.add(new Point(p1.x,p1.y));
    }
  }
}
