 /**----------------------------------------------------
 * File:      SVGTriangle.java
 * Package:   at.hoertler.svgeditor.svg.elements
 * Project:   SVGEditor
 * ----------------------------------------------------
 * @author Philipp Hörtler BSc
 * @version 1.0.0
 * Date: 2013-03-24
 * ----------------------------------------------------
 * Description:
 * 
 * Represents an SVGElement.
 */
package at.hoertler.svgeditor.svg.elements;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.event.MouseEvent;
import java.util.Vector;
import javax.swing.Icon;

public class SVGTriangle
        extends SVGElement {
  
  public SVGTriangle() {
    this.setFillcolorSetable(true);
    this.getHandles().add(new Point(-1,-1));
    this.getHandles().add(new Point(-1,-1));
    this.getHandles().add(new Point(-1,-1));
    this.setTextSetable(false);
  }
  
  @Override
  public SVGElement getInstance() { 
    return new SVGTriangle();
  }

  @Override
  public void paint(Graphics g) { 
    this.paint(g, 1);
  }

  @Override
  public void paint(Graphics g, int zoomFactor) { 
    this.setGraphics((Graphics2D) g);
    int[] xa = new int[3];
    int[] ya = new int[3];
    for(int i = 0; i<this.getHandles().size(); i++) {
      Point p = this.calcZoom(this.getHandles().get(i), zoomFactor);
      xa[i] = p.x;
      ya[i] = p.y;
    }
    this.getGraphics().setStroke(getStroke(zoomFactor));
    if(this.isFillcolorSetable()) {
      this.getGraphics().setColor(this.getFillColor());
      this.getGraphics().fillPolygon(xa, ya, this.getHandles().size());
    }
    if(this.getLineStrength()!=-1) {
      this.getGraphics().setColor(this.getLineColor());
      this.getGraphics().drawPolygon(xa, ya, this.getHandles().size());
    }
  }

  @Override
  public boolean hit(Rectangle rect) { 
    int[] xa = new int[3];
    int[] ya = new int[3];
    for(int i = 0; i<this.getHandles().size(); i++) {
      xa[i] = this.getHandles().get(i).x;
      ya[i] = this.getHandles().get(i).y;
    }
    Polygon poly = new Polygon(xa,ya,this.getHandles().size());
    return poly.intersects(rect);
  }

  @Override
  public Icon getIcon() { 
    return new Icon() {
      public int getIconHeight() {
        return 16;
      }

      public int getIconWidth() {
        return 16;
      }

      @Override
      public void paintIcon(Component c, Graphics g, int x, int y) {
        
        Graphics2D g2 = (Graphics2D) g;
        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                            RenderingHints.VALUE_ANTIALIAS_ON);
        g2.setColor(Color.BLACK);
        g2.setStroke(new BasicStroke(1, BasicStroke.CAP_BUTT,
                                     BasicStroke.JOIN_MITER));

        int[] xa = {x + 8,
                    x + 3,
                    x + 13
        };
        int[] ya = {y + 2,
                    y + 13,
                    y + 13
        };
        g2.drawPolygon(xa, ya, 3);
      }
    };
  }

  @Override
  public Point getTranslationPointForHandle(int handleIndex) { 
    /*
     *              /
     *          0  /
     *            o 
     *          /   \
     *     ----o-----o  1
     *          2     \
     *                 \
     */
    return this.getHandles().get((handleIndex+2)%3);
  }

  @Override
  public void moveHandle(int handleIndex, Point p, MouseEvent e) { 
    if (this.isNormalMode()) return;

    this.getHandles().get(handleIndex).translate(p.x, p.y);
    this.updateBoundingRect(this.getHandles());
    System.out.println(exportToSVG());
  }

  @Override
  public boolean validate() { 
    for(int i = 0; i<this.getHandles().size(); i++) {
      if(this.getHandles().get(i).distance(this.getHandles().get((i+1)%3)) == 0 ||
         this.getHandles().get(i).distance(this.getHandles().get((i+2)%3)) == 0)
        return false;
    }
    return true;
  }

  @Override
  public String exportToSVG() { 
    Vector<Point> elems = this.getHandles();
    StringBuilder sb = new StringBuilder();
    Color col = this.getLineColor();
    sb.append("\t");
    sb.append("<g id=\"" + this.getClass().getSimpleName() + "\">\n");
    sb.append("\t\t");
    sb.append("<path ");
    sb.append("d=\"M");
    sb.append(elems.firstElement().x + "," + elems.firstElement().y);
    for(int i = 1; i<elems.size(); i++)
      sb.append("L" + elems.get(i).x + "," + elems.get(i).y);
    sb.append("Z");

    sb.append("\" ");
    sb.append("stroke=\"rgb(" + col.getRed() + "," + col.getGreen() + "," + col.
            getBlue() + ")\" ");
    Color fc = this.getFillColor();
    if (fc.getAlpha() != 0)
      sb.append("fill=\"rgb(" + fc.getRed() + "," + fc.getGreen() + "," + fc.
              getBlue() + ")\" ");
    else
      sb.append("fill=\"none\" ");
    sb.append("stroke-width=\"" + this.getExportLineStrength() + "\"");
    sb.append(" />");
    sb.append("\n");
    sb.append("\t");
    sb.append("</g>\n");

    return sb.toString();
  }

  @Override
  public void handleMousePressed(MouseEvent e) { 
    if(!this.isInitialCreateMode()) return;
    
    this.getHandles().set(0, e.getPoint());
    this.getHandles().set(1, e.getPoint());
    this.getHandles().set(2, e.getPoint());
    this.updateBoundingRect(this.getHandles());
  }

  @Override
  public void handleMouseDragged(MouseEvent e) { 
    if(!this.isInitialCreateMode()) return;
    
    this.getHandles().set(1, e.getPoint());
    Point p2 = (Point)e.getPoint().clone();
    p2.x -= (e.getPoint().x - this.getHandles().get(0).x)*2;
    this.getHandles().set(2, p2);
    this.updateBoundingRect(this.getHandles());
  }

  @Override
  public void handleMouseReleased(MouseEvent e) { 
    if(!this.isInitialCreateMode()) return;

    this.updateBoundingRect(this.getHandles());
    this.setMode(Mode.Edit);
  }

  @Override
  public void handleMouseClicked(MouseEvent e) { }
}
